from collections.abc import Iterable, Iterator
from itertools import chain
from typing import Deque, Optional, TypeVar, Union

import numpy as np

from ..util.seeding import RngType

ExpType = TypeVar("ExpType")


class ExperienceReplay(Deque[ExpType]):
    """Deque-based class for RL traning to save and sample experience transitions. The
    class inherits from `collections.deque`, adding a couple of simple functionalities
    to it for sampling transitions at random from past observed data."""

    def __init__(
        self,
        iterable: Iterable[ExpType] = (),
        maxlen: Optional[int] = None,
        sample_size: Union[int, float] = 1,
        include_latest: Union[int, float] = 0,
        seed: RngType = None,
    ) -> None:
        """Instantiate the container for experience replay memory.

        Parameters
        ----------
        iterable : Iterable of T, optional
            Initial items to be inserted in the container. By default, empty.
        maxlen : int, optional
            Maximum length/capacity of the memory. If `None`, the deque has no maximum
            size, which is the default behaviour.
        sample_size : int or float, optional
            Size (as integer, or float percentage of `maxlen`) of the experience replay
            items to draw when performing an update. By default, one item per sampling
            is drawn. If a float percentage, `maxlen` must be provided.
        include_latest : int or float, optional
            Size (as integer, or float percentage of `sample_size`) dedicated to
            including the latest experience items. By default, 0, i.e., no last item is
            included.
        seed : None, int, array_like[ints], SeedSequence, BitGenerator, Generator
            Seed for the random number generator. By default, `None`.

        Raises
        ------
        TypeError
            Raises if `sample_size` is a float (a percentage of the maximum length), but
            `maxlen` is `None`, since it is impossible to compute the percentage of an
            unknown quantity.
        """
        if isinstance(sample_size, float) and maxlen is None:
            raise TypeError(
                "Cannot compute the percentage of an unknown quantity (maxlen is None)."
            )
        super().__init__(iterable, maxlen=maxlen)
        self.sample_size = sample_size
        self.include_latest = include_latest
        self.reset(seed)

    def reset(self, seed: RngType = None) -> None:
        """Resets the sampling RNG."""
        self.np_random = np.random.default_rng(seed)

    def sample(self) -> Iterator[ExpType]:
        """Samples the experience memory and yields the sampled items.

        Returns
        -------
        sample : iterator of T
            An iterable sample is yielded.
        """
        L = len(self)
        n = self.sample_size
        last_n = self.include_latest
        if isinstance(n, float):
            n = int(self.maxlen * n)
        n = min(max(n, 0), L)
        if isinstance(last_n, float):
            last_n = int(n * last_n)
        last_n = min(max(last_n, 0), n)

        # get last n indices and the sampled indices from the remaining
        last = range(L - last_n, L)
        sampled = self.np_random.choice(range(L - last_n), n - last_n, False)
        yield from (self[i] for i in chain(sampled, last))



from abc import ABC
from typing import Any, Generic, TypeVar

from ...optim.gradient_based_optimizer import GradientBasedOptimizer, LrType
from .agent import SymType
from .learning_agent import LearningAgent

ExpType = TypeVar("ExpType")


class RlLearningAgent(
    LearningAgent[SymType, ExpType], ABC, Generic[SymType, ExpType, LrType]
):
    """Base class for learning agents that employe gradient-based RL strategies to
    learn/improve the MPC policy."""

    def __init__(
        self,
        discount_factor: float,
        optimizer: GradientBasedOptimizer[SymType, LrType],
        **kwargs: Any,
    ) -> None:
        """Instantiates the RL learning agent.

        Parameters
        ----------
        discount_factor : float
            In RL, the factor that discounts future rewards in favor of immediate
            rewards. Usually denoted as `\\gamma`. Should be a number in (0, 1).
        optimizer : GradientBasedOptimizer
            A gradient-based optimizer (e.g., `mpcrl.optim.GradientDescent`) to compute
            the updates of the learnable parameters, based on the current gradient-based
            RL algorithm.
        kwargs
            Additional arguments to be passed to `LearningAgent`.
        """
        self.discount_factor = discount_factor
        self.optimizer = optimizer
        super().__init__(**kwargs)
        self.optimizer.set_learnable_parameters(self._learnable_pars)

    def establish_callback_hooks(self) -> None:
        super().establish_callback_hooks()
        optim = self.optimizer
        optimizer_hook = optim.hook
        if optimizer_hook is not None:
            self.hook_callback(repr(optim), optimizer_hook, optim.step)



__version__ = "1.2.1rc2"

__all__ = [
    "Agent",
    "ExperienceReplay",
    "GlobOptLearningAgent",
    "LearnableParameter",
    "LearnableParametersDict",
    "LearningAgent",
    "LstdDpgAgent",
    "LstdQLearningAgent",
    "MpcSolverError",
    "MpcSolverWarning",
    "RlLearningAgent",
    "UpdateError",
    "UpdateStrategy",
    "UpdateWarning",
    "WarmStartStrategy",
    "exploration",
    "optim",
    "schedulers",
    "wrappers_agents",
    "wrappers_envs",
]

from . import optim
from .agents.common.agent import Agent
from .agents.common.globopt_learning_agent import GlobOptLearningAgent
from .agents.common.learning_agent import LearningAgent
from .agents.common.rl_learning_agent import RlLearningAgent
from .agents.lstd_dpg import LstdDpgAgent
from .agents.lstd_q_learning import LstdQLearningAgent
from .core import exploration, schedulers
from .core.errors import MpcSolverError, MpcSolverWarning, UpdateError, UpdateWarning
from .core.experience import ExperienceReplay
from .core.parameters import LearnableParameter, LearnableParametersDict
from .core.update import UpdateStrategy
from .core.warmstart import WarmStartStrategy
from .wrappers import agents as wrappers_agents
from .wrappers import envs as wrappers_envs



from typing import Callable, Optional, TypeVar

import numpy as np
import numpy.typing as npt
from scipy.linalg import solve_discrete_are

T = TypeVar("T")


def dlqr(
    A: npt.NDArray[np.floating],
    B: npt.NDArray[np.floating],
    Q: npt.NDArray[np.floating],
    R: npt.NDArray[np.floating],
    M: Optional[npt.NDArray[np.floating]] = None,
) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]:
    """Get the discrete-time LQR for the given system. Stage costs are
    ```
        x'Qx + 2*x'Mu + u'Ru
    ```
    with `M = 0`, if not provided.

    Parameters
    ----------
    A : array
        State matrix.
    B : array
        Control input matrix.
    Q : array
        State weighting matrix.
    R : array
        Control input weighting matrix.
    M : array, optional
        Mixed state-input weighting matrix, by default None.

    Returns
    -------
    tuple of two arrays
        Returns the optimal state feedback matrix `K` and the quadratic terminal
        cost-to-go matrix `P`.

    Note
    ----
    Inspired by
    https://bitbucket.org/rawlings-group/mpc-tools-casadi/src/master/mpctools/util.py.
    """
    if M is not None:
        RinvMT = np.linalg.solve(R, M.T)
        Atilde = A - B.dot(RinvMT)
        Qtilde = Q - M.dot(RinvMT)
    else:
        Atilde = A
        Qtilde = Q
        M = np.zeros(B.shape)
    P = solve_discrete_are(Atilde, B, Qtilde, R)
    K = np.linalg.solve(B.T.dot(P).dot(B) + R, B.T.dot(P).dot(A) + M.T)
    return K, P


def rk4(f: Callable[[T], T], x0: T, dt: float = 1, M: int = 1) -> T:
    """Computes the Runge-Kutta 4 integration of the given function `f` with initial
    state x0.

    Parameters
    ----------
    f : Callable[[casadi or array], casadi or array]
        A function that takes a state as input and returns the derivative of the state,
        i.e., continuous-time dynamics.
    x0 : casadi or array
        The initial state. Must be compatible as argument to `f`.
    dt : float, optional
        The discretization timestep, by default `1`.
    M : int, optional
        How many RK4 steps to take in one `dt` interval, by default `1`.

    Returns
    -------
    new state : casadi or array
        The new state after `dt` time, according to the discretization.

    Note
    ----
    Inspired by
    https://bitbucket.org/rawlings-group/mpc-tools-casadi/src/master/mpctools/util.py.
    """
    dt /= M
    x = x0
    for _ in range(M):
        k1 = f(x)
        k2 = f(x + k1 * dt / 2)  # type: ignore
        k3 = f(x + k2 * dt / 2)  # type: ignore
        k4 = f(x + k3 * dt)  # type: ignore
        x = x + (k1 + 2 * k2 + 2 * k3 + k4) * dt / 6  # type: ignore
    return x



from typing import Any, Callable, Literal, Optional, TypeVar, Union

import numpy as np
import numpy.typing as npt
from gymnasium import Env

from .errors import raise_or_warn_on_mpc_failure, raise_or_warn_on_update_failure

ObsType = TypeVar("ObsType")
ActType = TypeVar("ActType")


def _failure_msg(
    category: Literal["mpc", "update"],
    name: str,
    episode: int,
    timestep: Optional[int],
    status: str,
) -> str:
    """Internal utility for composing message for mpc/update failure."""
    C = category.title()
    if timestep is None:
        return f"{C} failure of {name} at episode {episode}, status: {status}."
    else:
        return (
            f"{C} failure of {name} at episode {episode}, time {timestep}, "
            f"status: {status}."
        )


class CallbackMixin:
    """A class with the particular purpose of creating, storing and deleting hooks.
    Particularly touchy is when the state is set, and the hooks need to be reestablished
    automatically. In fact, if the old hooks are used, the new object (created from the
    state) would reference callbacks belonging to the old agent. In this way, the
    callbacks are linked to the new instance.
    """

    def __init__(self) -> None:
        self._hooks: dict[str, dict[str, Callable[..., None]]] = {}

    def __setstate__(
        self,
        state: Union[
            None, dict[str, Any], tuple[Optional[dict[str, Any]], dict[str, Any]]
        ],
    ) -> None:
        if isinstance(state, tuple) and len(state) == 2:
            state, slotstate = state
        else:
            slotstate = None
        if state is not None:
            # remove hooks (otherwise, new copies will still be calling the old object)
            state["_hooks"] = {}
            self.__dict__.update(state)
        if slotstate is not None:
            for key, value in slotstate.items():
                setattr(self, key, value)
        # re-establish hooks
        self.establish_callback_hooks()

    def _run_hooks(self, method_name: str, *args: Any) -> None:
        """Runs the internal hooks attached to the given method."""
        if hooks := self._hooks.get(method_name):
            for hook in hooks.values():
                hook(*args)

    def establish_callback_hooks(self) -> None:
        """This method must be used to perform the connections between callbacks and any
        invokable method (hook). If the object has no hooks, then this method does
        nothing."""

    def hook_callback(
        self, attachername: str, callbackname: str, func: Callable[..., None]
    ) -> None:
        """Hooks a function to be called each time a callback is invoked.

        Parameters
        ----------
        attachername : str
            The name of the object requesting the hook. Has only info purposes.
        callbackname : str
            Name of the callback to hook to, i.e., the target of the hooking.
        func : Callable
            function to be called when the callback is invoked. Must accept the same
            input arguments as the callback it is hooked to. Moreover, the return value
            is discarded.

        Raises
        ------
        ValueError
            If an hook with name `attachername` is already attached to this callback.
        """
        hook_dict = self._hooks.setdefault(callbackname, {})
        if attachername in hook_dict:
            raise ValueError(
                f"Hook '{attachername}' already attached to callback '{callbackname}'."
            )
        hook_dict[attachername] = func


class AgentCallbackMixin(CallbackMixin):
    """Callbacks for agents."""

    def on_mpc_failure(
        self, episode: int, timestep: Optional[int], status: str, raises: bool
    ) -> None:
        """Callback in case of MPC failure.

        Parameters
        ----------
        episode : int
            Number of the episode when the failure happened.
        timestep : int or None
            Timestep of the current episode when the failure happened. Can be `None` in
            case the error occurs inter-episodically.
        status : str
            Status of the solver that failed.
        raises : bool
            Whether the failure should be raised as exception.
        """
        name: str = getattr(self, "name", "agent")
        raise_or_warn_on_mpc_failure(
            _failure_msg("mpc", name, episode, timestep, status),
            raises,
        )
        self._run_hooks("on_mpc_failure", episode, timestep, status, raises)

    def on_validation_start(self, env: Env[ObsType, ActType]) -> None:
        """Callback called at the beginning of the validation process.

        Parameters
        ----------
        env : gym env
            A gym environment where the agent is being validated on.
        """
        self._run_hooks("on_validation_start", env)

    def on_validation_end(
        self, env: Env[ObsType, ActType], returns: npt.NDArray[np.floating]
    ) -> None:
        """Callback called at the end of the validation process.

        Parameters
        ----------
        env : gym env
            A gym environment where the agent has been validated on.
        returns : array of double
            Each episode's cumulative rewards.
        """
        self._run_hooks("on_validation_end", env, returns)

    def on_episode_start(
        self, env: Env[ObsType, ActType], episode: int, state: ObsType
    ) -> None:
        """Callback called at the beginning of each episode in the training process.

        Parameters
        ----------
        env : gym env
            A gym environment where the agent is being trained on.
        episode : int
            Number of the training episode.
        state : ObsType
            Starting state for this episode.
        """
        self._run_hooks("on_episode_start", env, episode, state)

    def on_episode_end(
        self, env: Env[ObsType, ActType], episode: int, rewards: float
    ) -> None:
        """Callback called at the end of each episode in the training process.

        Parameters
        ----------
        env : gym env
            A gym environment where the agent is being trained on.
        episode : int
            Number of the training episode.
        rewards : float
            Cumulative rewards for this episode.
        """
        self._run_hooks("on_episode_end", env, episode, rewards)

    def on_env_step(
        self, env: Env[ObsType, ActType], episode: int, timestep: int
    ) -> None:
        """Callback called after each `env.step`.

        Parameters
        ----------
        env : gym env
            A gym environment where the agent is being trained on.
        episode : int
            Number of the training episode.
        timestep : int
            Time instant of the current training episode.
        """
        self._run_hooks("on_env_step", env, episode, timestep)

    def on_timestep_end(
        self, env: Env[ObsType, ActType], episode: int, timestep: int
    ) -> None:
        """Callback called at the end of each time iteration. It is called with the same
        frequency as `env.step`, but with different timing.

        Parameters
        ----------
        env : gym env
            A gym environment where the agent is being trained on.
        episode : int
            Number of the training episode.
        timestep : int
            Time instant of the current training episode.
        """
        self._run_hooks("on_timestep_end", env, episode, timestep)


class LearningAgentCallbackMixin(AgentCallbackMixin):
    """Callbacks for learning agents."""

    def on_update_failure(
        self, episode: int, timestep: Optional[int], errormsg: str, raises: bool
    ) -> None:
        """Callback in case of update failure.

        Parameters
        ----------
        episode : int
            Number of the episode when the failure happened.
        timestep : int or None
            Timestep of the current episode when the failure happened. Can be `None` in
            case the update occurs inter-episodically.
        errormsg : str
            Error message of the update failure.
        raises : bool
            Whether the failure should be raised as exception.
        """
        name: str = getattr(self, "name", "agent")
        raise_or_warn_on_update_failure(
            _failure_msg("update", name, episode, timestep, errormsg),
            raises,
        )
        self._run_hooks("on_update_failure", episode, timestep, errormsg, raises)

    def on_training_start(self, env: Env[ObsType, ActType]) -> None:
        """Callback called at the beginning of the training process.

        Parameters
        ----------
        env : gym env
            A gym environment where the agent is being trained on.
        """
        self._run_hooks("on_training_start", env)

    def on_training_end(
        self, env: Env[ObsType, ActType], returns: npt.NDArray[np.floating]
    ) -> None:
        """Callback called at the end of the training process.

        Parameters
        ----------
        env : gym env
            A gym environment where the agent has been trained on.
        returns : array of double
            Each episode's cumulative rewards.
        """
        self._run_hooks("on_training_end", env, returns)

    def on_update(self) -> None:
        """Callback called after each `agent.update`. Use this callback for, e.g.,
        decaying exploration probabilities or learning rates."""
        self._run_hooks("on_update")



__all__ = [
    "Adam",
    "GradientBasedOptimizer",
    "GradientDescent",
    "GradientFreeOptimizer",
    "GD",
    "NetwonMethod",
    "NM",
    "RMSprop",
]

from .adam import Adam
from .gradient_based_optimizer import GradientBasedOptimizer
from .gradient_descent import GradientDescent
from .gradient_free_optimizer import GradientFreeOptimizer
from .newton_method import NetwonMethod
from .rmsprop import RMSprop

GD = GradientDescent
NM = NetwonMethod



from collections.abc import Iterator
from typing import Literal

from ..util.iters import bool_cycle, chain, repeat


class UpdateStrategy:
    """A class for the update strategy."""

    def __init__(
        self,
        frequency: int,
        hook: Literal["on_episode_end", "on_timestep_end"] = "on_timestep_end",
        skip_first: int = 0,
    ) -> None:
        """Initializes the update strategy.

        Parameters
        ----------
        frequency : int
            Frequency at which, each time the hook is called, an update should be
            carried out.
        skip_first : int, optional
            Skips the first `skip_first` updates. By default 0, so no update is skipped.
            This is useful when, e.g., the agent has to wait for the experience buffer
            to be filled before starting to update.
        hook : {'on_episode_end', 'on_timestep_end'}, optional
            Specifies to which callback to hook, i.e., when to check if an update is due
            according to the given frequency. The options are:
             - `on_episode_end` checks for an update after each episode's end
             - `on_timestep_end` checks for an update after each env's timestep.

            By default, 'on_timestep_end' is selected.
        """
        self.frequency = frequency
        self.hook = hook
        self._update_cycle = chain(
            repeat(False, skip_first * frequency), bool_cycle(frequency)
        )

    def can_update(self) -> bool:
        """Returns whether an update must be carried out at the current instant
        according to the strategy.

        Notes
        -----
        This methods steps internal iterators to check whether an update is due, so
        calling this method again will return a different value.

        Returns
        -------
        bool
            `True` if the agent should update according to this strategy; otherwise,
            `False`.
        """
        return next(self._update_cycle)

    def __iter__(self) -> Iterator[bool]:
        """With `__next__`, makes this class act like an iterator."""
        return self._update_cycle

    def __next__(self) -> bool:
        """With `__iter__`, makes this class act like an iterator."""
        return next(self._update_cycle)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(frequency={self.frequency},hook={self.hook})"

    def __str__(self) -> str:
        return self.__repr__()



from collections.abc import Collection, Iterator
from typing import Callable, Generic, Literal, Optional, SupportsFloat, Union

import casadi as cs
import numba as nb
import numpy as np
import numpy.typing as npt
from csnlp.wrappers import Mpc, NlpSensitivity
from gymnasium import Env
from typing_extensions import TypeAlias

from ..core.experience import ExperienceReplay
from ..core.exploration import ExplorationStrategy, NoExploration
from ..core.parameters import LearnableParametersDict
from ..core.update import UpdateStrategy
from ..core.warmstart import WarmStartStrategy
from ..optim.gradient_based_optimizer import GradientBasedOptimizer
from ..util.math import monomials_basis_function
from .common.agent import ActType, ObsType, SymType
from .common.rl_learning_agent import LrType, RlLearningAgent

ExpType: TypeAlias = tuple[
    npt.NDArray[np.floating],  # rollout's costs
    npt.NDArray[np.floating],  # rollout's state feature vectors Phi(s)
    npt.NDArray[np.floating],  # rollout's Psi(s,a)
    npt.NDArray[np.floating],  # rollout's gradient of policy w.r.t. theta
    npt.NDArray[np.floating],  # rollout's CAFA weight v
]


class LstdDpgAgent(RlLearningAgent[SymType, ExpType, LrType], Generic[SymType, LrType]):
    """Least-Squares Temporal Difference (LSTD) Deterministic Policy Gradient (DPG)
    agent, as first introduced in [1] as its stochastic counterpart, and then refined in
    [2]. An application can be found in [3].

    The DPG agent uses an MPC controller as policy provider and function approximation,
    and adjusts its parametrization according to the temporal-difference error, with the
    goal of improving the policy, in a direct fashion by estimating the gradient of the
    policy and descending in its direction.

    References
    ----------
    [1] Gros, S. and Zanon, M., 2021, May. Reinforcement Learning based on MPC
        and the Stochastic Policy Gradient Method. In 2021 American Control
        Conference (ACC) (pp. 1947-1952). IEEE.
    [2] Gros, S. and Zanon, M., 2019. Towards Safe Reinforcement Learning Using NMPC and
        Policy Gradients: Part II - Deterministic Case. arXiv preprint arXiv:1906.04034.
    [3] Cai, W., Kordabad, A.B., Esfahani, H.N., Lekkas, A.M. and Gros, S., 2021,
        December. MPC-based reinforcement learning for a simplified freight mission of
        autonomous surface vehicles. In 2021 60th IEEE Conference on Decision and
        Control (CDC) (pp. 2990-2995). IEEE.
    """

    def __init__(
        self,
        mpc: Mpc[SymType],
        update_strategy: Union[int, UpdateStrategy],
        discount_factor: float,
        optimizer: GradientBasedOptimizer,
        learnable_parameters: LearnableParametersDict[SymType],
        exploration: Optional[ExplorationStrategy],
        fixed_parameters: Union[
            None, dict[str, npt.ArrayLike], Collection[dict[str, npt.ArrayLike]]
        ] = None,
        experience: Union[None, int, ExperienceReplay[ExpType]] = None,
        warmstart: Union[
            Literal["last", "last-successful"], WarmStartStrategy
        ] = "last-successful",
        hessian_type: Literal["none", "natural"] = "none",
        rollout_length: int = -1,
        record_policy_performance: bool = False,
        record_policy_gradient: bool = False,
        state_features: Optional[cs.Function] = None,
        linsolver: Literal["csparse", "mldivide"] = "csparse",
        ridge_regression_regularization: float = 1e-6,
        use_last_action_on_fail: bool = False,
        name: Optional[str] = None,
    ) -> None:
        """Instantiates the LSTD DPG agent.

        Parameters
        ----------
        mpc : Mpc[casadi.SX or MX]
            The MPC controller used as policy provider by this agent. The instance is
            modified in place to create the approximations of the state function `V(s)`
            and action value function `Q(s,a)`, so it is recommended not to modify it
            further after initialization of the agent. Moreover, some parameter and
            constraint names will need to be created, so an error is thrown if these
            names are already in use in the mpc. These names are under the attributes
            `perturbation_parameter`, `action_parameter` and `action_constraint`.
        update_strategy : UpdateStrategy or int
            The strategy used to decide which frequency to update the mpc parameters
            with. If an `int` is passed, then the default strategy that updates every
            `n` episodes is used (where `n` is the argument passed); otherwise, an
            instance of `UpdateStrategy` can be passed to specify these in more details.
        discount_factor : float
            In RL, the factor that discounts future rewards in favor of immediate
            rewards. Usually denoted as `\\gamma`. Should be a number in (0, 1).
        optimizer : GradientBasedOptimizer
            A gradient-based optimizer (e.g., `mpcrl.optim.GradientDescent`) to compute
            the updates of the learnable parameters, based on the current gradient-based
            RL algorithm.
        learnable_parameters : LearnableParametersDict
            A special dict containing the learnable parameters of the MPC, together with
            their bounds and values. This dict is complementary with `fixed_parameters`,
            which contains the MPC parameters that are not learnt by the agent.
        exploration : ExplorationStrategy, optional
            Exploration strategy for inducing exploration in the online MPC policy (it
            is mandatory to explore in DPG).
        fixed_parameters : dict[str, array_like] or collection of, optional
            A dict (or collection of dict, in case of `csnlp.MultistartNlp`) whose keys
            are the names of the MPC parameters and the values are their corresponding
            values. Use this to specify fixed parameters, that is, non-learnable. If
            `None`, then no fixed parameter is assumed.
        experience : int or ExperienceReplay, optional
            The container for experience replay memory. If `None` is passed, then a
            memory with length 1 is created, i.e., it keeps only the latest memory
            transition. If an integer `n` is passed, then a memory with the length `n`
            is created and with sample size `n`.
            In case of LSTD DPG, each memory item is obtain from a single rollout, and
            is a 4-tuple that contains: costs, state vector features (Phi), Psi (a
            temporary value), and the gradient of the policy.
        max_percentage_update : float, optional
            A positive float that specifies the maximum percentage the parameters can be
            changed during each update. For example, `max_percentage_update=0.5` means
            that the parameters can be updated by up to 50% of their current value. By
            default, it is set to `+inf`.
        weight_decay : float, optional
            A positive float that specifies the decay of the learnable parameters in the
            form of an L2 regularization term. By default, it is set to `0.0`, so no
            decay/regularization takes place.
        warmstart: "last" or "last-successful" or WarmStartStrategy, optional
            The warmstart strategy for the MPC's NLP. If `last-successful`, the last
            successful solution is used to warm start the solver for the next iteration.
            If `last`, the last solution is used, regardless of success or failure.
            Furthermoer, a `WarmStartStrategy` object can be passed to specify a
            strategy for generating multiple warmstart points for the NLP. This is
            useful to generate multiple initial conditions for very non-convex problems.
            Can only be used with an MPC that has an underlying multistart NLP problem
            (see `csnlp.MultistartNlp`).
        hessian_type : {'none', 'natural'}, optional
            The type of hessian to use in this (potentially) second-order algorithm.
            If 'none', no second order information is used. If `natural`, the Fisher
            information matrix is used to perform a natural policy gradient update. This
            option must be in accordance with the choice of `optimizer`, that is, if the
            optimizer does not use second order information, then this option must be
            set to `none`.
        rollout_length : int, optional
            Number of steps of each closed-loop simulation, which defines a complete
            trajectory of the states (i.e., a rollout), and is saved in the experience
            as a single item (since LSTD DPG needs to draw samples of trajectories). In
            case the env is episodic, it can be `-1`, in which case the rollout length
            coincides with the episode's length. In case the env is not episodic, i.e.,
            it never terminates, a length `>0` must be given in order to know when to
            save the current trajectory as an atomic item in memory.
        record_policy_performance: bool, optional
            If `True`, the performance of each rollout is stored in the field
            `policy_performances`, which otherwise is `None`. By default, does not
            record them.
        record_policy_gradient: bool, optional
            If `True`, the (estimated) policy gradient of each update is stored in the
            field `policy_gradients`, which otherwise is `None`. By default, does not
            record them.
        state_features : casadi.Function, optional
            The state feature vector to be used in the linear approximation of the
            value function, which takes the form of
            ```
            V_v(s) = Phi(s)^T * v
            ```
            where `s` is the state, `v` are the weights, and `Phi(s)` is the state
            feature vector. This function is assumed to have one input and one output.
            By default, if not provided, it is designed as all monomials of the state
            with degrees <= 2 (see `mpcrl.util.math.monomials_basis_function`).
        linsolver : "csparse" or "mldivide", optional
            The type of linear solver to be used for solving the linear system derived
            from the KKT conditions and used to estimate the gradient of the policy. By
            default, `"csparse"` is chosen as the KKT matrix is most often sparse.
        ridge_regression_regularization : float, optional
            Ridge regression regularization used during the computations of the LSTD
            weights via least-squares. By default, `1e-6`.
        use_last_action_on_fail : bool, optional
            In case the MPC solver fails
             * if `False`, the action from the last solver's iteration is returned
               anyway (though suboptimal)
             * if `True`, the action from the last successful call to the MPC is
               returned instead (if the MPC has been solved at least once successfully).

            By default, `False`.
        name : str, optional
            Name of the agent. If `None`, one is automatically created from a counter of
            the class' instancies.

        Raises
        ------
        ValueError
            If the exploration strategy is `None` or an instance of `NoExploration`, as
            DPG requires exploration.
        """
        if exploration is None or isinstance(exploration, NoExploration):
            raise ValueError("DPG requires exploration, but none was provided.")
        # change default update hook to 'on_episode_end'
        if not isinstance(update_strategy, UpdateStrategy):
            update_strategy = UpdateStrategy(update_strategy, "on_episode_end")
        super().__init__(
            mpc=mpc,
            update_strategy=update_strategy,
            discount_factor=discount_factor,
            optimizer=optimizer,
            learnable_parameters=learnable_parameters,
            fixed_parameters=fixed_parameters,
            exploration=exploration,
            experience=experience,
            warmstart=warmstart,
            use_last_action_on_fail=use_last_action_on_fail,
            name=name,
        )
        self.hessian_type = hessian_type
        self._sensitivity = self._init_sensitivity(linsolver)
        self._Phi = (
            monomials_basis_function(mpc.ns, 0, 2)
            if state_features is None
            else state_features
        )
        self.regularization = ridge_regression_regularization
        self.rollout_length = rollout_length
        self._rollout: list[
            tuple[
                ObsType,
                ActType,
                SupportsFloat,
                ObsType,
                npt.NDArray[np.floating],
            ]
        ] = nb.typed.List()
        self.policy_performances: Optional[list[float]] = (
            [] if record_policy_performance else None
        )
        self.policy_gradients: Optional[list[npt.NDArray[np.floating]]] = (
            [] if record_policy_gradient else None
        )

    def update(self) -> Optional[str]:
        sample = self.experience.sample()
        if self.hessian_type == "natural":
            dJdtheta, fisher_hessian = _estimate_gradient_update(
                sample, self.discount_factor, self.regularization, True
            )
        else:
            dJdtheta = _estimate_gradient_update(
                sample, self.discount_factor, self.regularization, False
            )
            fisher_hessian = None
        if self.policy_gradients is not None:
            self.policy_gradients.append(dJdtheta)
        return self.optimizer.update(dJdtheta, fisher_hessian)

    def train_one_episode(
        self,
        env: Env[ObsType, ActType],
        episode: int,
        init_state: ObsType,
        raises: bool = True,
    ) -> float:
        truncated = terminated = False
        timestep = 0
        rewards = 0.0
        state = init_state
        rollout_length = self.rollout_length
        action_space = getattr(env, "action_space", None)
        gradient_based_exploration = self.exploration.mode == "gradient-based"
        action_keys = self.V.actions.keys()

        while not (truncated or terminated):
            # compute V(s)
            action, sol = self.state_value(state, False, action_space=action_space)
            if gradient_based_exploration:
                # NOTE: if the exploration affects the gradient, unfortunately we have
                # to solve again the NLP (but deterministically this time)
                action_opt, sol_opt = self.state_value(
                    state, True, action_space=action_space
                )
            else:
                # otherwise, just retrieve manually the unperturbed optimal action
                action_opt = cs.vertcat(*(sol.vals[u][:, 0] for u in action_keys))
                sol_opt = sol

            # step the system with the action just computed
            state_new, cost, truncated, terminated, _ = env.step(action)
            self.on_env_step(env, episode, timestep)

            # store transition in experience
            if sol.success and sol_opt.success:
                # NOTE: according to Cai et al. [3], the sensitivities should naively be
                # computed with the solution of unpertubed MPC (i.e., sol_opt).
                # According to Gros and Zanon [2], it is hinted that the perturbed
                # solution should be used instead (sol).
                exploration = np.asarray((action - action_opt).elements())
                sol_vals = np.asarray(sol_opt.all_vals.elements())
                self._rollout.append((state, exploration, cost, state_new, sol_vals))
            else:
                status = f"{sol.status}/{sol_opt.status}"
                self.on_mpc_failure(episode, timestep, status, raises)

            # increase counters
            state = state_new
            rewards += float(cost)
            timestep += 1

            # first, check if current rollout has reached its length, and only then
            # invoke on_timestep_end (as it might trigger an update)
            if rollout_length > 0 and len(self._rollout) >= rollout_length:
                self._consolidate_rollout_into_memory()
            self.on_timestep_end(env, episode, timestep)

        # consolidate rollout at the end of episode, if no length was specified
        if rollout_length <= 0:
            self._consolidate_rollout_into_memory()
        return rewards

    def _init_sensitivity(self, linsolver: str) -> Callable[[cs.DM, int], np.ndarray]:
        """Internal utility to compute the derivatives w.r.t. the learnable parameters
        and other functions in order to estimate the policy gradient."""
        assert (self.hessian_type == "none" and self.optimizer._order == 1) or (
            self.hessian_type == "natural" and self.optimizer._order == 2
        ), "expected 1st-order (2nd-order) optimizer with `none` (`natural`) hessian"
        nlp = self._V.nlp
        y = nlp.primal_dual
        theta = cs.vvcat(self._learnable_pars.sym.values())
        u0 = cs.vcat(self._V.first_actions.values())
        x_lam_p = cs.vertcat(nlp.primal_dual, nlp.p)

        # compute first bunch of derivatives
        nlp_ = NlpSensitivity(nlp, theta)
        Kt = nlp_.jacobians["K-p"]
        Ky = nlp_.jacobians["K-y"]
        dydu0 = cs.jacobian_sparsity(u0, y).T

        # instantiate linear solver (must be MX, so SX has to be converted)
        if nlp.sym_type is cs.SX:
            x_lam_p, x_lam_p_sx = cs.MX.sym("in", *x_lam_p.shape), x_lam_p
            Kt, Ky = cs.Function("sx2mx", (x_lam_p_sx,), (Kt, Ky))(x_lam_p)
        solver = (
            cs.solve  # cs.mldivide
            if linsolver == "mldivide"
            else cs.Linsol("linsolver", linsolver, Ky.sparsity()).solve
        )

        # compute sensitivity and convert to function (faster runtime)
        dpidtheta = -solver(Ky, Kt).T @ dydu0
        sensitivity = cs.Function(
            "dpidtheta",
            (x_lam_p,),
            (dpidtheta,),
            ("x_lam_p",),
            ("dpidtheta",),
            {"cse": True},
        )
        ntheta, na = dpidtheta.shape

        # wrap to conveniently return arrays. Casadi does not support tensors with
        # >2 dims, so dpidtheta gets squished in the 3rd dim and needs reshaping
        def func(sol_values: cs.DM, N: int) -> np.ndarray:
            return (
                np.ascontiguousarray(sensitivity(sol_values.T).elements())
                .reshape(ntheta, na, N, order="F")
                .transpose((2, 0, 1))
            )

        return func

    def _consolidate_rollout_into_memory(self) -> None:
        """Internal utility to compact current rollout into a single item in memory."""
        # convert rollout to arrays and clear it
        N, S, E, L, vals = _consolidate_rollout(self._rollout, self._V.ns, self._V.na)
        self._rollout.clear()

        # compute Phi, dpidtheta, Psi, and CAFA weight v
        Phi = np.ascontiguousarray(self._Phi(S.T).elements()).reshape(N + 1, -1)
        dpidtheta = self._sensitivity(vals, N)
        Psi = (dpidtheta @ E).reshape(N, dpidtheta.shape[1])
        v = _compute_cafa_weight_v(Phi, L, self.discount_factor, self.regularization)

        # save to experience
        self.store_experience((L, Phi, Psi, dpidtheta, v))
        if self.policy_performances is not None:
            self.policy_performances.append(L.sum())


@nb.njit(cache=True, nogil=True, parallel=True)
def _consolidate_rollout(
    rollout: list[tuple[ObsType, ActType, float, ObsType, np.ndarray]],
    ns: int,
    na: int,
) -> tuple[int, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Internal utility to convert a rollout list to arrays."""
    N = len(rollout)
    S = np.empty((N + 1, ns))
    E = np.empty((N, na, 1))  # additional dim required for Psi
    L = np.empty(N)
    sol_vals = np.empty((N, rollout[0][-1].size))
    for i in nb.prange(N):
        s, e, cost, _, sol_val = rollout[i]
        S[i] = s.reshape(-1)
        E[i] = e
        L[i] = cost
        sol_vals[i] = sol_val
    S[-1] = rollout[-1][3].reshape(-1)
    return N, S, E, L, sol_vals


@nb.njit(cache=True, nogil=True)
def _compute_cafa_weight_v(
    Phi_all: np.ndarray, L: np.ndarray, discount_factor: float, regularization: float
) -> np.ndarray:
    """Compute the CAFA weight v via ridge regression."""
    # solve for v via ridge regression: -phi'(gamma phi+ - phi) v = phi'L
    Phi = Phi_all[:-1]
    Phi_next = Phi_all[1:]
    Phi_diff = discount_factor * Phi_next - Phi
    M = Phi_diff.T @ Phi @ Phi.T
    R = regularization * np.eye(M.shape[0])
    return np.linalg.solve(M @ Phi_diff + R, -M @ L)


@nb.njit(cache=True, nogil=True)
def _compute_cafa_weight_w(
    Phi_all: np.ndarray,
    Psi: np.ndarray,
    L: np.ndarray,
    v: np.ndarray,
    discount_factor: float,
    regularization: float,
) -> np.ndarray:
    """Compute the CAFA weight w via ridge regression."""
    # solve for w via ridge regression: psi' psi w = psi' (L + (gamma phi+ - phi) v)
    Phi = Phi_all[:-1]
    Phi_next = Phi_all[1:]
    Phi_diff = discount_factor * Phi_next - Phi
    A = Psi.T @ Psi
    b = Psi.T @ (L + Phi_diff @ v)
    R = regularization * np.eye(A.shape[0])
    return np.linalg.solve(A.T @ A + R, A.T @ b)


def _estimate_gradient_update(
    sample: Iterator[ExpType],
    discount_factor: float,
    regularization: float,
    return_fisher_hessian: bool,
) -> Union[np.ndarray, tuple[np.ndarray, np.ndarray]]:
    """Internal utility to estimate the gradient of the policy and possibly the
    Fisher information matrix as well."""
    # compute average v and w
    sample_ = list(sample)  # load whole iterator into a list
    v = np.mean([o[4] for o in sample_], 0)
    w_ = [
        _compute_cafa_weight_w(Phi, Psi, L, v, discount_factor, regularization)
        for L, Phi, Psi, _, _ in sample_
    ]
    w = np.mean(w_, 0)

    if return_fisher_hessian:
        # compute both policy gradient and Fisher information matrix
        fisher_hess_ = []
        dJdtheta_ = []
        for _, _, _, dpidtheta, _ in sample_:
            F = (dpidtheta @ dpidtheta.transpose((0, 2, 1))).sum(0)
            fisher_hess_.append(F)
            dJdtheta_.append(F @ w)
        return np.mean(dJdtheta_, 0), np.mean(fisher_hess_, 0)

    # compute only policy gradient estimate
    dJdtheta_ = [(o[3] @ o[3].transpose((0, 2, 1))).sum(0) @ w for o in sample_]
    return np.mean(dJdtheta_, 0)



from typing import Literal, Optional, Union

import casadi as cs
import numpy as np
import numpy.typing as npt

from ..core.parameters import LearnableParametersDict, SymType
from ..core.schedulers import Scheduler
from .gradient_based_optimizer import GradientBasedOptimizer, LrType


class RMSprop(GradientBasedOptimizer[SymType, LrType]):
    """RMSprop optimizer, based on [1,2].

    References
    ----------
    [1] Geoffrey Hinton, Nitish Srivastava, and Kevin Swersky. Neural networks for
        machine learning lecture 6a overview of mini-batch gradient descent. page 14,
        2012.
    [2] RMSprop - PyTorch 2.1 documentation.
        https://pytorch.org/docs/stable/generated/torch.optim.RMSprop.html
    """

    _order = 1
    _hessian_sparsity = "diag"
    """In RMSprop, hessian is at most diagonal, i.e., in case we have constraints."""

    def __init__(
        self,
        learning_rate: Union[LrType, Scheduler[LrType]],
        alpha: float = 0.99,
        eps: float = 1e-8,
        weight_decay: float = 0.0,
        momentum: float = 0.0,
        centered: bool = False,
        hook: Literal["on_update", "on_episode_end", "on_timestep_end"] = "on_update",
        max_percentage_update: float = float("+inf"),
    ) -> None:
        """Instantiates the optimizer.

        Parameters
        ----------
        learning_rate : float/array, scheduler
            The learning rate of the optimizer. A float/array can be passed in case the
            learning rate must stay constant; otherwise, a scheduler can be passed which
            will be stepped `on_update` by default (see `hook` argument).
        alpha : float, optional
            A positive float that specifies the decay rate of the running average of the
            gradient. By default, it is set to `0.99`.
        eps : float, optional
            Term added to the denominator to improve numerical stability. By default, it
            is set to `1e-8`.
        weight_decay : float, optional
            A positive float that specifies the decay of the learnable parameters in the
            form of an L2 regularization term. By default, it is set to `0.0`, so no
            decay/regularization takes place.
        momentum : float, optional
            A positive float that specifies the momentum factor. By default, it is set
            to `0.0`, so no momentum is used.
        centered : bool, optional
            If `True`, compute the centered RMSProp, i.e., the gradient is normalized by
            an estimation of its variance.
        hook : {'on_update', 'on_episode_end', 'on_timestep_end'}, optional
            Specifies when to step the optimizer's learning rate's scheduler to decay
            its value (see `step` method also). This allows to vary the rate over the
            learning iterations. The options are:
             - `on_update` steps the learning rate after each agent's update
             - `on_episode_end` steps the learning rate after each episode's end
             - `on_timestep_end` steps the learning rate after each env's timestep.

            By default, 'on_update' is selected.
        max_percentage_update : float, optional
            A positive float that specifies the maximum percentage change the learnable
            parameters can experience in each update. For example,
            `max_percentage_update=0.5` means that the parameters can be updated by up
            to 50% of their current value. By default, it is set to `+inf`. If
            specified, the update becomes constrained and has to be solved as a QP,
            which is inevitably slower than its unconstrained counterpart.
        """
        super().__init__(learning_rate, hook, max_percentage_update)
        self.weight_decay = weight_decay
        self.alpha = alpha
        self.eps = eps
        self.momentum = momentum
        self.centered = centered

    def set_learnable_parameters(self, pars: LearnableParametersDict[SymType]) -> None:
        super().set_learnable_parameters(pars)
        # initialize also running averages
        n = pars.size
        self._square_avg = np.zeros(n, dtype=float)
        self._grad_avg = np.zeros(n, dtype=float) if self.centered else None
        self._momentum_buf = np.zeros(n, dtype=float) if self.momentum > 0.0 else None

    def _first_order_update(
        self, gradient: npt.NDArray[np.floating]
    ) -> tuple[npt.NDArray[np.floating], Optional[str]]:
        theta = self.learnable_parameters.value

        # compute candidate update
        weight_decay = self.weight_decay
        lr = self.lr_scheduler.value
        if weight_decay > 0.0:
            gradient = gradient + weight_decay * theta
        dtheta, self._square_avg, self._grad_avg, self._momentum_buf = _rmsprop(
            gradient,
            self._square_avg,
            lr,
            self.alpha,
            self.eps,
            self.centered,
            self._grad_avg,
            self.momentum,
            self._momentum_buf,
        )

        # if unconstrained, apply the update directly; otherwise, solve the QP
        solver = self._update_solver
        if solver is None:
            return theta + dtheta, None
        lbx, ubx = self._get_update_bounds(theta)
        sol = solver(h=cs.DM.eye(theta.shape[0]), g=-dtheta, lbx=lbx, ubx=ubx)
        dtheta = np.asarray(sol["x"].elements())
        stats = solver.stats()
        return theta + dtheta, None if stats["success"] else stats["return_status"]


def _rmsprop(
    grad: np.ndarray,
    square_avg: np.ndarray,
    lr: LrType,
    alpha: float,
    eps: float,
    centered: bool,
    grad_avg: Optional[np.ndarray],
    momentum: float,
    momentum_buffer: Optional[np.ndarray],
) -> tuple[np.ndarray, np.ndarray, Optional[np.ndarray], Optional[np.ndarray]]:
    """Computes the update's change according to Adam algorithm."""
    square_avg = alpha * square_avg + (1 - alpha) * np.square(grad)

    if centered:
        grad_avg = alpha * grad_avg + (1 - alpha) * grad
        avg = np.sqrt(square_avg - np.square(grad_avg))
    else:
        avg = np.sqrt(square_avg)
    avg += eps

    if momentum > 0.0:
        momentum_buffer = momentum * momentum_buffer + grad / avg
        dtheta = -lr * momentum_buffer
    else:
        dtheta = -lr * grad / avg
    return dtheta, square_avg, grad_avg, momentum_buffer



from typing import Any, Literal, Optional, Union

import numpy as np
import numpy.typing as npt
from scipy.linalg import cho_solve

from ..core.schedulers import Scheduler
from ..util.math import cholesky_added_multiple_identities
from .gradient_based_optimizer import GradientBasedOptimizer, LrType, SymType


class NetwonMethod(GradientBasedOptimizer[SymType, LrType]):
    """Netwon Method."""

    _order = 2
    _hessian_sparsity = "dense"

    def __init__(
        self,
        learning_rate: Union[LrType, Scheduler[LrType]],
        weight_decay: float = 0.0,
        cho_before_update: bool = False,
        cho_maxiter: int = 1000,
        cho_solve_kwargs: Optional[dict[str, Any]] = None,
        hook: Literal["on_update", "on_episode_end", "on_timestep_end"] = "on_update",
        max_percentage_update: float = float("+inf"),
    ) -> None:
        """Instantiates the optimizer.

        Parameters
        ----------
        learning_rate : float/array, scheduler
            The learning rate of the optimizer. A float/array can be passed in case the
            learning rate must stay constant; otherwise, a scheduler can be passed which
            will be stepped `on_update` by default (see `hook` argument).
        weight_decay : float, optional
            A positive float that specifies the decay of the learnable parameters in the
            form of an L2 regularization term. By default, it is set to `0.0`, so no
            decay/regularization takes place.
        cho_before_update : bool, optional
            Whether to perform a Cholesky's factorization of the hessian in preparation
            of each update. If `False`, the QP update's objective is
            ```math
                min 1/2 * dtheta' * H * dtheta + (lr * g)' * dtheta
            ```
            else, if `True`, the objective is
            ```math
                min 1/2 * ||dtheta||^2' + (lr * H^-1 * g)' * dtheta
            ```
            where the hessian linear system is performed via Cholesky's factorization.
            Only relevant if the update is  constrained. By default, `False`.
        cho_maxiter : int, optional
            Maximum number of iterations in the Cholesky's factorization with additive
            multiples of the identity to ensure positive definiteness of the hessian. By
            default, `1000`.
        cho_solve_kwargs : kwargs for `scipy.linalg.cho_solve`, optional
            The optional kwargs to be passed to `scipy.linalg.cho_solve` to solve linear
            systems with the hessian's Cholesky decomposition. If `None`, it is
            equivalent to `cho_solve_kwargs = {'check_finite': False }`.
        hook : {'on_update', 'on_episode_end', 'on_timestep_end'}, optional
            Specifies when to step the optimizer's learning rate's scheduler to decay
            its value (see `step` method also). This allows to vary the rate over the
            learning iterations. The options are:
             - `on_update` steps the learning rate after each agent's update
             - `on_episode_end` steps the learning rate after each episode's end
             - `on_timestep_end` steps the learning rate after each env's timestep.

            By default, 'on_update' is selected.
        max_percentage_update : float, optional
            A positive float that specifies the maximum percentage change the learnable
            parameters can experience in each update. For example,
            `max_percentage_update=0.5` means that the parameters can be updated by up
            to 50% of their current value. By default, it is set to `+inf`. If
            specified, the update becomes constrained and has to be solved as a QP,
            which is inevitably slower than its unconstrained counterpart.
        """
        if cho_before_update:
            self._hessian_sparsity = "diag"
        super().__init__(learning_rate, hook, max_percentage_update)
        self.weight_decay = weight_decay
        self.cho_before_update = cho_before_update
        self.cho_maxiter = cho_maxiter
        if cho_solve_kwargs is None:
            cho_solve_kwargs = {"check_finite": False}
        self.cho_solve_kwargs = cho_solve_kwargs

    def _second_order_update(
        self, gradient: npt.NDArray[np.floating], hessian: npt.NDArray[np.floating]
    ) -> tuple[npt.NDArray[np.floating], Optional[str]]:
        theta = self.learnable_parameters.value
        lr = self.lr_scheduler.value
        w = self.weight_decay
        cho_kw = self.cho_solve_kwargs
        L = cholesky_added_multiple_identities(hessian, maxiter=self.cho_maxiter)

        # if unconstrained, apply the update directly; otherwise, solve the QP
        solver = self._update_solver
        if solver is None:
            dtheta = _nm_unconstrained(theta, gradient, L, lr, w, cho_kw)
            return theta + dtheta, None
        H, G = _nm_constrained(
            theta, gradient, hessian, L, lr, w, self.cho_before_update, cho_kw
        )
        lbx, ubx = self._get_update_bounds(theta)
        sol = solver(h=H, g=G, lbx=lbx, ubx=ubx)
        dtheta = np.asarray(sol["x"].elements())
        stats = solver.stats()
        return theta + dtheta, None if stats["success"] else stats["return_status"]


def _nm_unconstrained(
    theta: np.ndarray,
    g: np.ndarray,
    L: np.ndarray,
    lr: LrType,
    weight_decay: float,
    cho_solve_kwargs: dict,
) -> np.ndarray:
    """Computes the update's change according to the Netwon's Method."""
    if weight_decay <= 0.0:
        return -lr * cho_solve((L, True), g, **cho_solve_kwargs)
    return -np.linalg.solve(
        L @ L.T + weight_decay * np.eye(theta.shape[0]), lr * g + weight_decay * theta
    )


def _nm_constrained(
    theta: np.ndarray,
    g: np.ndarray,
    H: np.ndarray,
    L: np.ndarray,
    lr: LrType,
    weight_decay: float,
    cho_before_update: bool,
    cho_solve_kwargs: dict,
) -> tuple[np.ndarray, np.ndarray]:
    """Computes the update's change according to the Netwon's Method, which is solved
    numerically due to the presence of constraints."""
    if weight_decay <= 0.0:
        if cho_before_update:
            G = cho_solve((L, True), lr * g, **cho_solve_kwargs)
            H = np.eye(theta.shape[0])
        else:
            G = lr * g
            H = L @ L.T
    else:
        G = lr * g + weight_decay * theta
        H = L @ L.T + weight_decay * np.eye(theta.shape[0])
        if cho_before_update:
            G = np.linalg.solve(H, G)
            H = np.eye(theta.shape[0])
    return H, G



__all__ = ["Evaluate", "LearningWrapper", "Log", "RecordUpdates", "Wrapper"]

from .evaluate import Evaluate
from .log import Log
from .record_updates import RecordUpdates
from .wrapper import LearningWrapper, Wrapper



import numpy as np
import numpy.typing as npt

from ...agents.common.learning_agent import ExpType, LearningAgent
from .wrapper import LearningWrapper, SymType


class RecordUpdates(LearningWrapper[SymType, ExpType]):
    """Wrapper for recording the history of updates by the learning agent."""

    def __init__(self, agent: LearningAgent[SymType, ExpType]) -> None:
        """Instantiates the recorder.

        Parameters
        ----------
        agent : LearningAgent or inheriting
            The agent whose updates need recording.
        """
        super().__init__(agent)
        self.updates_history: dict[str, list[npt.NDArray[np.floating]]] = {
            p.name: [p.value] for p in agent.learnable_parameters.values()
        }

    def _on_update(self, *_, **__) -> None:
        """Internal utility to store the current parameters in memory."""
        for par in self.agent.learnable_parameters.values():
            self.updates_history[par.name].append(par.value)

    def establish_callback_hooks(self) -> None:
        super().establish_callback_hooks()
        # connect the agent's on_update callback to this wrapper storing action
        self.hook_callback(repr(self), "on_update", self._on_update)



from warnings import warn


class MpcSolverError(RuntimeError):
    """Exception class for raising errors when MPC solvers fails."""


class MpcSolverWarning(RuntimeWarning):
    """Warning class for raising errors when MPC solvers fails."""


class UpdateError(RuntimeError):
    """Exception class for raising errors when RL agent's updates fail."""


class UpdateWarning(RuntimeWarning):
    """Warning class for raising errors when RL agent's updates fail."""


def raise_or_warn_on_update_failure(msg: str, raises: bool) -> None:
    """Raises errors or warnings with a message for update failures.

    Parameters
    ----------
    msg : str
        The exception or warning message.
    raises : bool
        If `True`, raises an exception; otherwise, throws a warning.

    Raises
    ------
    UpdateError
        Raises `UpdateError` if `raises=True`; otherwise raises
        `UpdateWarning`.
    """
    if raises:
        raise UpdateError(msg)
    else:
        warn(msg, UpdateWarning)


def raise_or_warn_on_mpc_failure(msg: str, raises: bool) -> None:
    """Raises errors or warnings with a message for MPC failures.

    Parameters
    ----------
    msg : str
        The exception or warning message.
    raises : bool
        If `True`, raises an exception; otherwise, throws a warning.

    Raises
    ------
    UpdateError
        Raises `MpcSolverError` if `raises=True`; otherwise raises
        `MpcSolverWarning`.
    """
    if raises:
        raise MpcSolverError(msg)
    else:
        warn(msg, MpcSolverWarning)



from collections import deque
from time import perf_counter
from typing import Any, Deque, Optional, SupportsFloat, TypeVar

import numpy as np
import numpy.typing as npt
from gymnasium import Env, Wrapper, utils

ObsType = TypeVar("ObsType")
ActType = TypeVar("ActType")


class MonitorEpisodes(
    Wrapper[ObsType, ActType, ObsType, ActType], utils.RecordConstructorArgs
):
    """This wrapper keeps track of
        - observations (``observations``)
        - actions (``actions``)
        - costs/rewards (``rewards``)
        - episode length (``episode_lengths``)
        - episode execution time (``exec_times``)
    that the environment is subject to during the learning process. Note that these are
    effectly saved in each corresponding field only when the episode is done (terminated
    or truncated).

    After the completion of an episode, these fields will look like this::

        >>> env.observations = <deque of each episode's observations>
        ... env.actions = <deque of each episode's actions>
        ... env.rewards = <deque of each episode's rewards>
        ... env.episode_lengths = <deque of each episode's episode length>
        ... env.exec_times = <deque of each episode's execution time>
    """

    def __init__(
        self, env: Env[ObsType, ActType], deque_size: Optional[int] = None
    ) -> None:
        """This wrapper will keep track of observations, actions and rewards as well as
        episode length and execution time.

        Parameters
        ----------
        env : Env[ObsType, ActType]
            The environment to apply the wrapper to.
        deque_size : int, optional
            The maximum number of episodes to hold as historical data in the internal
            deques. By default, `None`, i.e., unlimited.
        """
        utils.RecordConstructorArgs.__init__(self, deque_size=deque_size)
        Wrapper.__init__(self, env)
        # long-term storages
        self.observations: Deque[npt.NDArray[ObsType]] = deque(  # type: ignore
            maxlen=deque_size
        )
        self.actions: Deque[npt.NDArray[ActType]] = deque(  # type: ignore
            maxlen=deque_size
        )
        self.rewards: Deque[npt.NDArray[np.floating]] = deque(maxlen=deque_size)
        self.episode_lengths: Deque[int] = deque(maxlen=deque_size)
        self.exec_times: Deque[float] = deque(maxlen=deque_size)
        # current-episode-storages
        self.ep_observations: list[ObsType] = []
        self.ep_actions: list[ActType] = []
        self.ep_rewards: list[SupportsFloat] = []
        self.t0: float = perf_counter()
        self.ep_length: int = 0

    def reset(
        self, *, seed: Optional[int] = None, options: Optional[dict[str, Any]] = None
    ) -> tuple[ObsType, dict[str, Any]]:
        """Resets the environment and resets the current data accumulators."""
        observation, info = super().reset(seed=seed, options=options)
        self._clear_ep_data()
        self.ep_observations.append(observation)
        return observation, info

    def step(
        self, action: ActType
    ) -> tuple[ObsType, SupportsFloat, bool, bool, dict[str, Any]]:
        # sourcery skip: extract-method
        """Steps through the environment, accumulating the episode data."""
        obs, reward, terminated, truncated, info = super().step(action)

        # accumulate data
        self.ep_observations.append(obs)
        self.ep_actions.append(action)
        self.ep_rewards.append(reward)
        self.ep_length += 1

        # if episode is done, save the current data to history
        if terminated or truncated:
            # append data
            self.observations.append(np.asarray(self.ep_observations))
            self.actions.append(np.asarray(self.ep_actions))
            self.rewards.append(np.asarray(self.ep_rewards))
            self.episode_lengths.append(self.ep_length)
            self.exec_times.append(perf_counter() - self.t0)

            # clear this episode's data
            self._clear_ep_data()

        return obs, reward, terminated, truncated, info

    def _clear_ep_data(self) -> None:
        # clear this episode's lists and reset counters
        self.ep_observations.clear()
        self.ep_actions.clear()
        self.ep_rewards.clear()
        self.t0 = perf_counter()
        self.ep_length = 0



import pickle
import unittest
from functools import lru_cache
from itertools import product
from unittest.mock import MagicMock, Mock, call
from warnings import catch_warnings

import casadi as cs
import numpy as np
from csnlp import Nlp, Solution, scaling
from csnlp.multistart import StackedMultistartNlp
from csnlp.wrappers import Mpc, NlpScaling
from parameterized import parameterized, parameterized_class
from scipy import io as matio

from mpcrl import (
    Agent,
    ExperienceReplay,
    LearnableParametersDict,
    LearningAgent,
    MpcSolverWarning,
)
from mpcrl import exploration as E
from mpcrl import schedulers as S
from mpcrl.util.seeding import mk_seed

OPTS = {
    "expand": True,
    "print_time": False,
    "ipopt": {
        "max_iter": 500,
        "sb": "yes",
        # for debugging
        "print_level": 0,
        "print_user_options": "no",
        "print_options_documentation": "no",
    },
}
RESULTS = matio.loadmat(r"tests/data_test_agent.mat")


@lru_cache
def get_dynamics(g: float, alpha: float, dt: float) -> cs.Function:
    x, u, d = cs.MX.sym("x", 3), cs.MX.sym("u", 2), cs.MX.sym("d", 3)
    x_next = x + cs.vertcat(x[1], u[0] / x[2] - g, -alpha * u[0]) * dt + d * 0
    return cs.Function("F", [x, u, d], [x_next], ["x", "u", "d"], ["x+"])


def get_mpc(horizon: int, multistart: bool):
    N = horizon
    T = 100
    K = 3
    dt = T / N
    yT = 100000
    g = 9.81
    alpha = 1 / (300 * g)
    y_nom = 1e5
    v_nom = 2e3
    m_nom = 3e5
    u_nom = 1e8
    scaler = scaling.Scaler()
    scaler.register("y", scale=y_nom)
    scaler.register("y_0", scale=y_nom)
    scaler.register("v", scale=v_nom)
    scaler.register("v_0", scale=v_nom)
    scaler.register("m", scale=m_nom)
    scaler.register("m_0", scale=m_nom)
    scaler.register("u1", scale=u_nom)
    scaler.register("u2", scale=u_nom)
    nlp = (
        StackedMultistartNlp[cs.MX](sym_type="MX", starts=K)
        if multistart
        else Nlp[cs.MX](sym_type="MX")
    )
    nlp = NlpScaling[cs.MX](nlp, scaler=scaler, warns=False)
    mpc = Mpc[cs.MX](nlp, prediction_horizon=N)
    y, _ = mpc.state("y")
    _, _ = mpc.state("v")
    m, _ = mpc.state("m", lb=0)
    mpc.action("u1", lb=0, ub=5e7)
    u2, _ = mpc.action("u2", lb=0, ub=5e7)
    mpc.disturbance("d", 3)
    mpc.set_dynamics(get_dynamics(g, alpha, dt))
    mpc.constraint("yT", y[-1], "==", yT)
    mpc.minimize(m[0] - m[-1] + cs.sum2(u2))
    mpc.init_solver(OPTS)
    return mpc


class DummyLearningAgent(LearningAgent):
    def update(self, *args, **kwargs):
        return

    def train_one_episode(self, *args, **kwargs):
        return


@parameterized_class("multistart_nlp", [(True,), (False,)])
class TestAgent(unittest.TestCase):
    def test_init__raises__mpc_with_no_actions(self):
        with self.assertRaisesRegex(
            ValueError, "Expected Mpc with na>0; got na=0 instead."
        ):
            Agent(mpc=Mpc(Nlp(), 4))

    def test_init__instantiates_V_and_Q_as_two_different_mpcs(self):
        agent = Agent(mpc=get_mpc(3, self.multistart_nlp))
        self.assertIsInstance(agent.Q, Mpc)
        self.assertIsInstance(agent.V, Mpc)
        self.assertIsNot(agent.Q, agent.V)

    @parameterized.expand([(None,), ("gradient-based",), ("additive",)])
    def test_init__instantiates_V_and_Q_correctly(self, exploration_mode: str):
        if exploration_mode is None:
            exploration = E.NoExploration()
        else:
            exploration = E.GreedyExploration(0.5, mode=exploration_mode)
        agent = Agent(mpc=get_mpc(3, self.multistart_nlp), exploration=exploration)

        V_pars_keys = agent.V.parameters.keys()
        Q_pars_keys = agent.Q.parameters.keys()
        if exploration_mode is None or exploration_mode == "additive":
            self.assertNotIn(agent.cost_perturbation_parameter, V_pars_keys)
        else:
            self.assertIn(agent.cost_perturbation_parameter, V_pars_keys)
        self.assertNotIn(agent.cost_perturbation_parameter, Q_pars_keys)
        self.assertIn(agent.init_action_parameter, Q_pars_keys)
        self.assertIn(agent.init_action_constraint, agent.Q.constraints.keys())
        self.assertNotIn(agent.init_action_parameter, V_pars_keys)
        self.assertNotIn(agent.init_action_constraint, agent.V.constraints.keys())

    def test_init__removes_bounds_on_initial_action_in_Q_correctly(self):
        mpc = get_mpc(3, self.multistart_nlp)
        mpc.unwrapped.remove_variable_bounds = Mock()
        agent = Agent(mpc=mpc, remove_bounds_on_initial_action=True)
        self.assertFalse(agent.V.remove_variable_bounds.called)
        calls = agent.Q.remove_variable_bounds.call_args_list
        self.assertEqual(len(calls), mpc.na)
        for i, mcall in enumerate(calls):
            expected_name = f"u{i + 1}"
            na_ = mpc.actions[expected_name].size1()
            name, direction, idx = mcall.args
            self.assertEqual(name, expected_name)
            self.assertEqual(direction, "both")
            self.assertEqual(list(idx), [(r, 0) for r in range(na_)])

    def test_unwrapped(self):
        agent = Agent(mpc=get_mpc(3, self.multistart_nlp))
        agent2 = agent.unwrapped
        self.assertIs(agent, agent2)

    @parameterized.expand([(True,), (False,)])
    def test__is_deepcopyable_and_pickleable(self, copy: bool):
        epsilon, epsilon_decay_rate = 0.7, 0.75
        strength, strength_decay_rate = 0.5, 0.75
        epsilon_scheduler = S.ExponentialScheduler(epsilon, epsilon_decay_rate)
        strength_scheduler = S.ExponentialScheduler(strength, strength_decay_rate)
        exploration = E.EpsilonGreedyExploration(
            epsilon=epsilon_scheduler, strength=strength_scheduler, seed=42
        )
        agent1 = Agent(mpc=get_mpc(3, self.multistart_nlp))
        agent1._exploration = exploration  # a bit of cheating

        if copy:
            agent2 = agent1.copy()
        else:
            agent2: Agent = pickle.loads(pickle.dumps(agent1))

        self.assertIsNot(agent1, agent2)
        self.assertIsNot(agent1.Q, agent2.Q)
        self.assertIsNot(agent1.V, agent2.V)
        self.assertIs(agent1.exploration, exploration)
        exp1, exp2 = agent1.exploration, agent2.exploration
        self.assertIsNot(exp1, exp2)
        self.assertIsNot(exp1, exp2)
        for scheduler in ["epsilon_scheduler", "strength_scheduler"]:
            sc1, sc2 = getattr(exp1, scheduler), getattr(exp2, scheduler)
            self.assertIsNot(sc1, sc2)
            self.assertEqual(sc1.value, sc2.value)
            self.assertEqual(sc1.factor, sc2.factor)

    @parameterized.expand(product(["V", "Q"], [False, True], [False, True]))
    def test_solve_mpc__calls_mpc_with_correct_args(
        self, mpctype: str, vector: bool, multiple_pars: bool
    ):
        # sourcery skip: low-code-quality
        starts = 3
        horizon = 3
        if not self.multistart_nlp:
            multiple_pars = False
        fixed_pars = {
            Agent.cost_perturbation_parameter: [42, 69],
            "d": cs.DM([5, 6, 7]),
        }
        fixed_pars_ = (
            (fixed_pars.copy() for _ in range(starts))
            if multiple_pars
            else fixed_pars.copy()
        )
        mpc = get_mpc(horizon, self.multistart_nlp)
        exploration = E.GreedyExploration(0.5, mode="gradient-based")
        agent = Agent(mpc=mpc, fixed_parameters=fixed_pars_, exploration=exploration)
        mpc: Mpc[cs.SX] = getattr(agent, mpctype)

        s = {"y": 0, "v": 10, "m": 5e5}
        a = {"u1": 1, "u2": 2}
        if vector:
            s = cs.DM(s.values())
            a = cs.DM(a.values())
        if mpctype == "V":
            a = None

        pert = cs.DM([65, 79])
        vals0 = object()
        sol = Solution(
            f=5,
            vars=None,
            vals=vals0,
            dual_vars=None,
            dual_vals=None,
            stats={"success": True},
            _get_value=None,
        )
        agent._last_solution = sol
        if self.multistart_nlp:
            mpc.nlp.solve_multi = MagicMock(return_value=sol)
        else:
            mpc.nlp.solve = MagicMock(return_value=sol)

        agent.solve_mpc(mpc, state=s, action=a, perturbation=pert)

        call_pars = {
            **fixed_pars,
            "y_0": s[0] if vector else s["y"],
            "v_0": s[1] if vector else s["v"],
            "m_0": s[2] if vector else s["m"],
        }
        if mpctype == "V":
            call_pars[Agent.cost_perturbation_parameter] = pert
        else:
            call_pars[Agent.init_action_parameter] = a if vector else cs.DM(a.values())
        if self.multistart_nlp:
            mpc.nlp.solve_multi.assert_called_once()
            call_args = mpc.nlp.solve_multi.call_args.args
        else:
            mpc.nlp.solve.assert_called_once()
            call_args = mpc.nlp.solve.call_args.args
        actual_call_pars, actual_call_vals0 = call_args
        if multiple_pars:
            for pars_i in actual_call_pars:
                self.assertEqual(len(mpc.unwrapped._pars.keys() - pars_i.keys()), 0)
                for key in call_pars:
                    np.testing.assert_allclose(pars_i[key], call_pars[key], rtol=0)
        else:
            pars = actual_call_pars
            self.assertEqual(len(mpc.unwrapped._pars.keys() - pars.keys()), 0)
            for key in call_pars:
                np.testing.assert_allclose(pars[key], call_pars[key], rtol=0)
        self.assertIs(actual_call_vals0, vals0)

    @parameterized.expand(product((False, True), (False, True)))
    def test_state_value__computes_right_solution(
        self, vector: bool, deterministic: bool
    ):  # sourcery skip: move-assign
        starts = 3
        horizon = 100
        fixed_pars = {"d": cs.DM([5, 6, 7])}
        if self.multistart_nlp:
            fixed_pars = [fixed_pars.copy() for _ in range(starts)]

        mpc = get_mpc(horizon, self.multistart_nlp)
        agent = Agent(mpc=mpc, fixed_parameters=fixed_pars)

        state = {"y": 0, "v": 0, "m": 5e5}
        vals0 = {"y": 0, "v": 0, "m": 5e5, "u1": 1e8, "u2": 0}
        agent._last_solution = Solution(
            f=0,
            vars=0,
            vals=vals0,
            dual_vars=None,
            dual_vals=None,
            stats=0,
            _get_value=0,
        )
        if vector:
            state = cs.DM(state.values())

        action, sol = agent.state_value(
            state=state, vals0=None, deterministic=deterministic
        )
        self.assertTrue(sol.success)
        np.testing.assert_array_equal(
            action, cs.vertcat(sol.vals["u1"][:, 0], sol.vals["u2"][:, 0])
        )
        np.testing.assert_allclose(sol.f, RESULTS["state_value_f"].item(), rtol=1e-3)
        np.testing.assert_allclose(
            sol.vals["u1"],
            RESULTS["state_value_us"],
            rtol=1e-7,
            atol=1e-7,
        )

    @parameterized.expand(product((False, True), (False, True)))
    def test_action_value__computes_right_solution(self, vector: bool, a_optimal: bool):
        # sourcery skip: move-assign
        starts = 3
        horizon = 100
        fixed_pars = {"d": cs.DM([5, 6, 7])}
        if self.multistart_nlp:
            fixed_pars = [fixed_pars.copy() for _ in range(starts)]

        mpc = get_mpc(horizon, self.multistart_nlp)
        agent = Agent(mpc=mpc, fixed_parameters=fixed_pars)

        a_opt, a_subopt = 5e7, 1.42e7
        state = {"y": 0, "v": 0, "m": 5e5}
        action = {"u1": a_opt if a_optimal else a_subopt, "u2": 0}
        vals0 = {**state, **action}
        agent._last_solution = Solution(
            f=0, vars=0, vals=vals0, dual_vars=0, dual_vals=0, stats=0, _get_value=0
        )
        if vector:
            state = cs.DM(state.values())
            action = cs.DM(action.values())

        sol = agent.action_value(state=state, action=action, vals0=None)
        self.assertTrue(sol.success)
        u1_0_0 = sol.value(mpc.unscaled_variables["u1"][0, 0])
        if a_optimal:
            np.testing.assert_allclose(u1_0_0, a_opt)
            np.testing.assert_allclose(
                sol.vals["u1"],
                RESULTS["state_value_us"],
                rtol=1e-7,
                atol=1e-7,
            )
        else:
            np.testing.assert_allclose(u1_0_0, a_subopt)
            self.assertTrue(sol.f >= RESULTS["state_value_f"].item())

    def test_evaluate__performs_correct_calls(self):
        seed = 69
        episodes = 3
        episode_length = 10
        Ttot = episodes * episode_length
        reset_states = [object() for _ in range(episodes)]
        reset_options = object()
        step_states = [object() for _ in range(Ttot)]
        rewards = np.random.randn(Ttot)
        truncateds = [False] * Ttot
        terminateds = np.full(Ttot, 0, dtype=bool)
        terminateds[::-episode_length] = True
        infos = [{}] * Ttot
        env = MagicMock()
        env.reset = Mock(side_effect=zip(reset_states, infos))
        env.step = Mock(
            side_effect=zip(step_states, rewards, truncateds, terminateds, infos)
        )
        actions1 = np.random.randn(Ttot, 2, 1)
        actions2 = np.random.randn(Ttot, 3, 1)
        actions = [cs.vertcat(u1, u2) for u1, u2 in zip(actions1, actions2)]
        successes = [True] * (Ttot - 1) + [False]
        sols = map(
            lambda u1, u2, success: Solution(
                0,
                {},
                {"u1": u1, "u2": u2},
                {},
                {},
                {"success": success, "return_status": "bad"},
                None,
            ),
            actions1,
            actions2,
            successes,
        )
        agent = Agent(mpc=get_mpc(3, False))
        agent.state_value = Mock(side_effect=zip(actions, sols))
        deterministic = object()

        with catch_warnings(record=True) as cw:
            returns = Agent.evaluate(
                agent,
                env,
                episodes,
                deterministic=deterministic,
                seed=seed,
                raises=False,
                env_reset_options=reset_options,
            )

        np.testing.assert_allclose(returns, rewards.reshape(-1, episode_length).sum(1))
        rng = np.random.default_rng(seed)
        env.reset.assert_has_calls(
            [call(seed=mk_seed(rng), options=reset_options) for i in range(episodes)]
        )
        for mcall, u1, u2 in zip(env.step.mock_calls, actions1, actions2):
            self.assertEqual(len(mcall.args), 1)
            np.testing.assert_array_equal(mcall.args[0], cs.vertcat(u1, u2))
        states = np.column_stack(
            (reset_states, np.asarray(step_states).reshape(-1, episode_length))
        )[:, :-1].flatten()
        agent.state_value.assert_has_calls(
            [call(state, deterministic) for state in states]
        )
        self.assertEqual(len(cw), 1)
        self.assertIs(cw[0].category, MpcSolverWarning)


class TestLearningAgent(unittest.TestCase):
    def test_init__creates_default_experience_and_exploration(self):
        learnable_parameters = LearnableParametersDict()
        agent = DummyLearningAgent(
            mpc=get_mpc(3, False),
            learnable_parameters=learnable_parameters,
            update_strategy=1,
        )
        self.assertIsInstance(agent.exploration, E.ExplorationStrategy)
        self.assertIsInstance(agent.experience, ExperienceReplay)
        self.assertIs(agent.learnable_parameters, learnable_parameters)

    def test_store_experience__appends_experience(self):
        experience = MagicMock()
        experience.append = Mock()
        item = object()
        agent = DummyLearningAgent(
            mpc=get_mpc(3, False),
            learnable_parameters={},
            experience=experience,
            update_strategy=1,
        )
        self.assertIs(agent.experience, experience)
        agent.store_experience(item)
        experience.append.assert_called_once_with(item)


if __name__ == "__main__":
    unittest.main()



from abc import ABC
from typing import Any, Generic, Optional

from gymnasium import Env

from mpcrl.agents.common.agent import ActType, ObsType

from ...core.update import UpdateStrategy
from ...optim.gradient_free_optimizer import GradientFreeOptimizer
from .agent import SymType
from .learning_agent import LearningAgent


class GlobOptLearningAgent(LearningAgent[SymType, None], ABC, Generic[SymType]):
    """Class for learning agents that employe gradient-free Global Optimization
    strategies (e.g., Bayesian Optimization) to learn/improve the MPC policy.

    An instance of this class requires a gradient-free optimizer that adheres to the
    ask-tell interface, i..e., it must implement the `ask` and `tell` methods.
    """

    def __init__(
        self, optimizer: GradientFreeOptimizer[SymType], **kwargs: Any
    ) -> None:
        """Instantiates the RL learning agent.

        Parameters
        ----------
        optimizer : GradientFreeOptimizer
            A gradient-free optimizer to ask for parameters and tell the values of the
            objective function.
        kwargs
            Additional arguments to be passed to `LearningAgent`.

            Note: the following kwargs are not yet supported
             - "experience": usually, GO strategies do not require experience replay
             - "update_strategy": updates are fixed at the end of each episode
        """
        for key in ("experience", "update_strategy"):
            if key in kwargs:
                raise ValueError(
                    f"{self.__class__.__name__} does not yet support `{key}` kwargs."
                )
        self.optimizer = optimizer
        super().__init__(update_strategy=UpdateStrategy(1, "on_episode_end"), **kwargs)
        self.optimizer.set_learnable_parameters(self._learnable_pars)

    def train(self, *args: Any, **kwargs: Any) -> Any:
        # prime the initial value of the learnable parameters by asking it to the
        # optimizer. This is the usual way to start Global Optimization strategies,
        # whereas in RL we usually start the user's given initial values.
        self.update()
        return super().train(*args, **kwargs)

    def train_one_episode(
        self,
        env: Env[ObsType, ActType],
        episode: int,
        init_state: ObsType,
        raises: bool = True,
    ) -> float:
        # simply evaluate the MPC on the env with the current set of parameters for one
        # episode, and then tell the optimizer the value of the objective function
        rewards = 0.0
        state = init_state
        truncated, terminated, timestep = False, False, 0

        while not (truncated or terminated):
            action, sol = self.state_value(state, False)
            if not sol.success:
                self.on_mpc_failure(episode, timestep, sol.status, raises)

            state, r, truncated, terminated, _ = env.step(action)
            self.on_env_step(env, episode, timestep)

            rewards += float(r)
            timestep += 1
            self.on_timestep_end(env, episode, timestep)

        values = (
            self._learnable_pars.value_as_dict
            if self.optimizer.prefers_dict
            else self._learnable_pars.value
        )
        self.optimizer.tell(values, rewards)
        return rewards

    def update(self) -> Optional[str]:
        # simply ask the optimizer for a new set of parameters, and then update the
        # current parameters with this new set
        theta_new, status = self.optimizer.ask()
        self._learnable_pars.update_values(theta_new)
        return status



from abc import ABC, abstractmethod
from collections.abc import Collection, Iterable
from typing import Any, Callable, Generic, Optional, TypeVar, Union

import numpy as np
import numpy.typing as npt
from gymnasium import Env
from gymnasium.spaces import Box

from ...core.callbacks import LearningAgentCallbackMixin
from ...core.experience import ExperienceReplay
from ...core.parameters import LearnableParametersDict
from ...core.update import UpdateStrategy
from ...util.seeding import RngType, mk_seed
from .agent import ActType, Agent, ObsType, SymType, _update_dicts

ExpType = TypeVar("ExpType")


class LearningAgent(
    Agent[SymType], LearningAgentCallbackMixin, ABC, Generic[SymType, ExpType]
):
    """Base class for a learning agent with MPC as policy provider where the main method
    `update`, which is called to update the learnable parameters of the MPC according to
    the underlying learning methodology (e.g., Bayesian Optimization, RL, etc.) is
    abstract and must be implemented by inheriting classes.

    Aside of `update`, this class implements also the basic structure for both on-policy
    and off-policy learning, which require further ad-hoc implementations. For on-policy
    learning, the `train` method should be run, which requires the implementation of
    `train_one_episode`. For off-policy learning, run `train_offpolicy` and implement
    `train_one_rollout` instead. At least one of the two implementations is required in
    the inheriting class, depending on the learning algorithm. Some algorithms might
    support both.

    Note: this class makes no assumptions on the learning methodology used to update the
    MPC's learnable parameters."""

    def __init__(
        self,
        update_strategy: Union[int, UpdateStrategy],
        learnable_parameters: LearnableParametersDict[SymType],
        experience: Union[None, int, ExperienceReplay[ExpType]] = None,
        **kwargs: Any,
    ) -> None:
        """Instantiates the learning agent.

        Parameters
        ----------
        update_strategy : UpdateStrategy or int
            The strategy used to decide which frequency to update the mpc parameters
            with. If an `int` is passed, then the default strategy that updates every
            `n` env's steps is used (where `n` is the argument passed); otherwise, an
            instance of `UpdateStrategy` can be passed to specify these in more details.
        learnable_parameters : LearnableParametersDict
            A special dict containing the learnable parameters of the MPC, together with
            their bounds and values. This dict is complementary with `fixed_parameters`,
            which contains the MPC parameters that are not learnt by the agent.
        experience : int or ExperienceReplay, optional
            The container for experience replay memory. If `None` is passed, then a
            memory with length 1 is created, i.e., it keeps only the latest memory
            transition. If an integer `n` is passed, then a memory with the length `n`
            is created and with sample size `n`.
        kwargs
            Additional arguments to be passed to `Agent`.
        """
        Agent.__init__(self, **kwargs)
        LearningAgentCallbackMixin.__init__(self)
        self._raises: bool = True
        self._learnable_pars = learnable_parameters
        if experience is None:
            experience = ExperienceReplay(maxlen=1)
        elif isinstance(experience, int):
            experience = ExperienceReplay(maxlen=experience, sample_size=experience)
        self._experience = experience
        if not isinstance(update_strategy, UpdateStrategy):
            update_strategy = UpdateStrategy(update_strategy, "on_timestep_end")
        self._update_strategy = update_strategy
        self._updates_enabled = True
        self.establish_callback_hooks()

    @property
    def experience(self) -> ExperienceReplay[ExpType]:
        """Gets the experience replay memory of the agent."""
        return self._experience

    @property
    def update_strategy(self) -> UpdateStrategy:
        """Gets the update strategy of the agent."""
        return self._update_strategy

    @property
    def learnable_parameters(self) -> LearnableParametersDict[SymType]:
        """Gets the parameters of the MPC that can be learnt by the agent."""
        return self._learnable_pars

    def reset(self, seed: RngType = None) -> None:
        """Resets agent's internal variables, exploration and experience's RNG"""
        super().reset(seed)
        self.experience.reset(seed)

    def store_experience(self, item: ExpType) -> None:
        """Stores the given item in the agent's memory for later experience replay.

        Parameters
        ----------
        item : experience-type
            Item to be stored in memory.
        """
        self._experience.append(item)

    def evaluate(self, *args: Any, **kwargs: Any) -> npt.NDArray[np.floating]:
        self._updates_enabled = False
        return super().evaluate(*args, **kwargs)

    def train(
        self,
        env: Env[ObsType, ActType],
        episodes: int,
        seed: RngType = None,
        raises: bool = True,
        env_reset_options: Optional[dict[str, Any]] = None,
    ) -> npt.NDArray[np.floating]:
        """On-policy training of the agent on an environment.

        Parameters
        ----------
        env : Env[ObsType, ActType]
            A gym environment where to train the agent in.
        episodes : int
            Number of training episodes.
        seed : None, int, array_like[ints], SeedSequence, BitGenerator, Generator
            Agent's and each env's RNG seed.
        raises : bool, optional
            If `True`, when any of the MPC solver runs fails, or when an update fails,
            the corresponding error is raised; otherwise, only a warning is raised.
        env_reset_options : dict, optional
            Additional information to specify how the environment is reset at each
            training episode (optional, depending on the specific environment).

        Returns
        -------
        array of doubles
            The cumulative returns for each training episode.

        Raises
        ------
        MpcSolverError or MpcSolverWarning
            Raises the error or the warning (depending on `raises`) if any of the MPC
            solvers fail.
        UpdateError or UpdateWarning
            Raises the error or the warning (depending on `raises`) if the update fails.
        """
        if hasattr(env, "action_space"):
            assert isinstance(env.action_space, Box), "Env action space must be a Box,"
        rng = np.random.default_rng(seed)
        self.reset(rng)
        self._updates_enabled = True
        self._raises = raises
        returns = np.zeros(episodes, float)

        self.on_training_start(env)
        for episode in range(episodes):
            state, _ = env.reset(seed=mk_seed(rng), options=env_reset_options)
            self.on_episode_start(env, episode, state)
            r = self.train_one_episode(env, episode, state, raises)
            self.on_episode_end(env, episode, r)
            returns[episode] = r

        self.on_training_end(env, returns)
        return returns

    def train_one_episode(
        self,
        env: Env[ObsType, ActType],
        episode: int,
        init_state: ObsType,
        raises: bool = True,
    ) -> float:
        """On-policy training of the agent on an environment for one episode.

        Parameters
        ----------
        env : Env[ObsType, ActType]
            A gym environment where to train the agent in.
        episode : int
            Number of the current training episode.
        init_state : observation type
            Initial state/observation of the environment.
        raises : bool, optional
            If `True`, when any of the MPC solver runs fails, or when an update fails,
            the corresponding error is raised; otherwise, only a warning is raised.

        Returns
        -------
        float
            The cumulative rewards for this training episode.

        Raises
        ------
        MpcSolverError or MpcSolverWarning
            Raises the error or the warning (depending on `raises`) if any of the MPC
            solvers fail.
        UpdateError or UpdateWarning
            Raises the error or the warning (depending on `raises`) if the update fails.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement `train_one_episode` for "
            "on-policy learning."
        )

    def train_offpolicy(
        self,
        episode_rollouts: Iterable[Iterable[Any]],
        seed: RngType = None,
        raises: bool = True,
    ) -> None:
        """Off-policy training of the agent on an environment.

        Parameters
        ----------
        episode_rollouts : iterable of iterables of any
            An iterable (i.e., a sequence) of episodical rollouts generated off-policy.
            Each rollout is itself a sequence of transitions, e.g., SARSA tuples. In
            general, these can be of different types, depending on the specific learning
            algorithm.
        seed : None, int, array_like[ints], SeedSequence, BitGenerator, Generator
            Agent's (and each evaluation env's, if applicable) RNG seed.
        raises : bool, optional
            If `True`, when any of the MPC solver runs fails, or when an update fails,
            the corresponding error is raised; otherwise, only a warning is raised.

        Raises
        ------
        MpcSolverError or MpcSolverWarning
            Raises the error or the warning (depending on `raises`) if any of the MPC
            solvers fail.
        UpdateError or UpdateWarning
            Raises the error or the warning (depending on `raises`) if the update fails.
        """
        rng = np.random.default_rng(seed)
        self.reset(rng)
        self._raises = raises
        env_proxy = "off-policy"

        self._updates_enabled = True
        self.on_training_start(env_proxy)
        for episode, rollout in enumerate(episode_rollouts):
            self.on_episode_start(env_proxy, episode, float("nan"))
            self.train_one_rollout(rollout, episode, raises)
            self.on_episode_end(env_proxy, episode, float("nan"))

        self.on_training_end(env_proxy, np.empty(0))

    def train_one_rollout(
        self, rollout: Iterable[Any], episode: int, raises: bool = True
    ) -> None:
        """Train the agent in an off-policy manner on the given rollout.

        Parameters
        ----------
        rollout : iterable of any
            Rollout, i.e., a sequence of transitions generated off-policy, e.g., SARSA
            tuples. In general, these can be of different types, depending on the
            specific learning algorithm.
        raises : bool, optional
            If `True`, when any of the MPC solver runs fails, or when an update fails,
            the corresponding error is raised; otherwise, only a warning is raised.

        Raises
        ------
        MpcSolverError or MpcSolverWarning
            Raises the error or the warning (depending on `raises`) if any of the MPC
            solvers fail.
        UpdateError or UpdateWarning
            Raises the error or the warning (depending on `raises`) if the update fails.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement `train_offpolicy` for "
            "off-policy learning."
        )

    @abstractmethod
    def update(self) -> Optional[str]:
        """Updates the learnable parameters of the MPC according to the agent's learning
        algorithm.

        Returns
        -------
        errormsg : str or None
            In case the update fails, an error message is returned to be raised as error
            or warning; otherwise, `None` is returned.
        """

    def establish_callback_hooks(self) -> None:
        super().establish_callback_hooks()
        # hook exploration (only if necessary)
        exploration_hook = self._exploration.hook
        if exploration_hook is not None:
            self.hook_callback(
                repr(self._exploration), exploration_hook, self._exploration.step
            )
        # hook updates (always necessary)
        update_hook = self._update_strategy.hook
        func: Callable[..., None] = (
            (lambda _, e, __: self._check_and_perform_update(e, None))
            if update_hook == "on_episode_end"
            else (lambda _, e, t: self._check_and_perform_update(e, t))
        )
        self.hook_callback(repr(self._update_strategy), update_hook, func)

    def _check_and_perform_update(self, episode: int, timestep: Optional[int]) -> None:
        """Internal utility to check if an update is due and perform it."""
        if not self._updates_enabled or not self._update_strategy.can_update():
            return
        update_msg = self.update()
        if update_msg is not None:
            self.on_update_failure(episode, timestep, update_msg, self._raises)
        self.on_update()

    def _get_parameters(
        self,
    ) -> Union[None, dict[str, npt.ArrayLike], Collection[dict[str, npt.ArrayLike]]]:
        """Internal utility to retrieve parameters of the MPC in order to solve it.
        `LearningAgent` returns both fixed and learnable parameters."""
        learnable_pars = self._learnable_pars.value_as_dict
        fixed_pars = self._fixed_pars
        if fixed_pars is None:
            return learnable_pars  # type: ignore[return-value]
        if isinstance(fixed_pars, dict):
            fixed_pars.update(learnable_pars)
            return fixed_pars
        return tuple(
            _update_dicts(self._fixed_pars, learnable_pars)  # type: ignore[arg-type]
        )



from collections.abc import Iterable
from functools import cached_property
from itertools import chain
from numbers import Integral
from typing import Any, Generic, Optional, TypeVar, Union

import numpy as np
import numpy.typing as npt
from csnlp.core.cache import invalidate_cache
from csnlp.util.io import SupportsDeepcopyAndPickle

from ..util.math import summarize_array

SymType = TypeVar("SymType")  # most likely, T is cs.SX or MX


class LearnableParameter(SupportsDeepcopyAndPickle, Generic[SymType]):
    """A parameter that is learnable, that is, it can be adjusted via RL. This class
    is useful for managing symbols, bounds and value of learnable parameters."""

    def __init__(
        self,
        name: str,
        shape: Union[int, tuple[int, ...]],
        value: npt.ArrayLike,
        lb: npt.ArrayLike = -np.inf,
        ub: npt.ArrayLike = +np.inf,
        sym: Optional[SymType] = None,
    ) -> None:
        """Instantiates a learnable parameter.

        Parameters
        ----------
        name : str
            Name of the learnable parameter.
        shape : int or tuple of ints
            Shape of the parameter.
        value : array_like
            Starting value of the parameter. This can then be updated via `update`.
        lb : array_like, optional
            Lower bound of the parameter values. If not specified, it is unbounded.
        ub : array_like, optional
            Upper bound of the parameter values. If not specified, it is unbounded.
        sym : T, optional
            An optional reference to a symbolic variable representing this parameter.

        Raises
        ------
        ValueError
            Raises if `value`, `lb` or `ub` cannot be broadcasted to a 1D vector with
            shape equal to `shape`; or if the shape of the symbolic variable `sym` does
            not match the shape of the parameter.
        """
        super().__init__()
        self.name = name
        self.shape: tuple[int, ...] = (
            (shape,) if isinstance(shape, Integral) else shape  # type: ignore
        )
        self.sym = sym
        self.lb: npt.NDArray[np.floating] = np.broadcast_to(lb, shape)
        self.ub: npt.NDArray[np.floating] = np.broadcast_to(ub, shape)
        self._check_sym_shape()
        self._update_value(value)

    @property
    def size(self) -> int:
        """Gets the number of elements in the parameter."""
        return np.prod(self.shape, dtype=int).item()

    def _update_value(self, v: npt.ArrayLike, **is_close_kwargs: Any) -> None:
        """Internal utility for updating the parameter value with a new value.

        Parameters
        ----------
        new_value : array_like
            New value of the parameter.
        is_close_kwargs
            Additional kwargs for `np.isclose`, e.g., `rtol` and `atol`, for checking
            numerical values close to a bound.

        Raises
        ------
        ValueError
            Raises if `new_value` cannot be broadcasted to a 1D vector with shape equal
            to `shape`; or if it does not lie inside the upper and lower bounds within
            the specified tolerances.
        """
        v = np.broadcast_to(v, self.shape)
        lb = self.lb
        ub = self.ub
        if ((v < lb) & ~np.isclose(v, lb, **is_close_kwargs)).any() or (
            (v > ub) & ~np.isclose(v, ub, **is_close_kwargs)
        ).any():
            raise ValueError(f"Updated parameter `{self.name}` is outside bounds.")
        self.value: npt.NDArray[np.floating] = np.clip(v, lb, ub)

    def _check_sym_shape(self) -> None:
        """Internal utility for checking that the shape of the symbolic variable matches
        the shape of the parameter."""
        if self.sym is None or not hasattr(self.sym, "shape"):
            return
        sym_shape = tuple(self.sym.shape)
        shape = self.shape + tuple(1 for _ in range(len(sym_shape) - len(self.shape)))
        if sym_shape != shape:
            raise ValueError("Shape of `sym` does not match `shape`.")

    def __str__(self) -> str:
        return f"<{self.name}(shape={self.shape})>"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(name={self.name},shape={self.shape})>"


class LearnableParametersDict(
    dict[str, LearnableParameter[SymType]], SupportsDeepcopyAndPickle
):
    """dict-based collection of `LearnableParameter` instances that simplifies the
    process of managing and updating these. The dict contains pairs of parameter's name
    vs parameter's instance.

    Note: to speed up computations, properties of this class are often cached for faster
    calls to the same methods. However, these are cleared when the underlying dict is
    modified."""

    def __init__(self, pars: Optional[Iterable[LearnableParameter[SymType]]] = None):
        """Initializes the collection of learnable parameters.

        Parameters
        ----------
        pars : iterable of LearnableParameter, optional
            An optional iterable of parameters to insert into the dict by their names.
        """
        if pars is None:
            dict.__init__(self)
        else:
            dict.__init__(self, map(lambda p: (p.name, p), chain(pars)))
        SupportsDeepcopyAndPickle.__init__(self)

    @cached_property
    def size(self) -> int:
        """Gets the overall size of all the learnable parameters."""
        return sum(p.size for p in self.values())

    @cached_property
    def lb(self) -> npt.NDArray[np.floating]:
        """Gets the lower bound of all the learnable parameters, concatenated."""
        return (
            np.concatenate([p.lb.reshape(-1, order="F") for p in self.values()])
            if self
            else np.empty(0)
        )

    @cached_property
    def ub(self) -> npt.NDArray[np.floating]:
        """Gets the upper bound of all the learnable parameters, concatenated."""
        return (
            np.concatenate([p.ub.reshape(-1, order="F") for p in self.values()])
            if self
            else np.empty(0)
        )

    @cached_property
    def value(self) -> npt.NDArray[np.floating]:
        """Gets the values of all the learnable parameters, concatenated."""
        return (
            np.concatenate([p.value.reshape(-1, order="F") for p in self.values()])
            if self
            else np.empty(0)
        )

    @cached_property
    def value_as_dict(self) -> dict[str, npt.NDArray[np.floating]]:
        """Gets the values of all the learnable parameters, in a dict."""
        return {p.name: p.value for p in self.values()}

    @cached_property
    def sym(self) -> dict[str, Optional[SymType]]:
        """Gets symbols of all the learnable parameters, in a dict. If one parameter
        does not possess the symbol, `None` is put."""
        return {
            parname: None if par.sym is None else par.sym
            for parname, par in self.items()
        }

    @invalidate_cache(value, value_as_dict)
    def update_values(
        self,
        new_values: Union[npt.ArrayLike, dict[str, npt.ArrayLike]],
        **is_close_kwargs: Any,
    ) -> None:
        """Updates the value of each parameter

        Parameters
        ----------
        new_values : array_like or dict[str, array_like]
            The parameters' new values, either as a single concatenated array (which
            will be splitted according to the sizes and each piece sequentially assigned
            to each parameter), or as a dict of parameter's name vs parameter's new
            value.
        is_close_kwargs
            Additional kwargs for `np.isclose`, e.g., `rtol` and `atol`, for checking
            numerical values of parameters close to a bound.

        Raises
        ------
        ValueError
            In case of array-like, raises if `new_values` cannot be split according to
            the sizes of parameters; or if the new values cannot be broadcasted to 1D
            vectors according to each parameter's size; or if the new values lie outside
            either the lower or upper bounds of each parameter.
        """
        if isinstance(new_values, dict):
            for parname, new_value in new_values.items():
                self[parname]._update_value(new_value, **is_close_kwargs)
        else:
            cumsizes = np.cumsum([p.size for p in self.values()])[:-1]
            values_ = np.split(new_values, cumsizes)
            for par, val in zip(self.values(), values_):
                par._update_value(val.reshape(par.shape, order="F"), **is_close_kwargs)

    __cache_decorator = invalidate_cache(size, lb, ub, value, value_as_dict, sym)

    @__cache_decorator
    def __setitem__(self, name: str, par: LearnableParameter[SymType]) -> None:
        assert name == par.name, f"Key '{name}' must match parameter name '{par.name}'."
        return super().__setitem__(name, par)

    @__cache_decorator
    def update(
        self,
        pars: Iterable[LearnableParameter[SymType]],
        *args: LearnableParameter[SymType],
    ) -> None:
        return super().update(map(lambda p: (p.name, p), chain(pars, args)))

    @__cache_decorator
    def setdefault(
        self,
        par: LearnableParameter[SymType],
    ) -> LearnableParameter[SymType]:
        return super().setdefault(par.name, par)

    __delitem__ = __cache_decorator(dict.__delitem__)
    pop = __cache_decorator(dict.pop)
    popitem = __cache_decorator(dict.popitem)
    clear = __cache_decorator(dict.clear)

    def copy(
        self, deep: bool = False, invalidate_caches: bool = True
    ) -> "LearnableParametersDict[SymType]":
        """Creates a shallow or deep copy of the dict of learnable parameters.

        Parameters
        ----------
        deep : bool, optional
            If `True`, a deepcopy of the dict and its parameters is returned; otherwise,
            the copy is only shallow.
        invalidate_caches : bool, optional
            If `True`, methods decorated with `csnlp.util.funcs.invalidate_cache` are
            called to clear cached properties/lru caches in the copied instance.
            Otherwise, caches in the copy are not invalidated. By default, `True`.
            Only relevant when `deep=True`.

        Returns
        -------
        LearnableParametersDict[T]
            A copy of the dict of learnable parameters.
        """
        return (
            SupportsDeepcopyAndPickle.copy(self, invalidate_caches)
            if deep
            else LearnableParametersDict[SymType](self.values())
        )

    def stringify(
        self, summarize: bool = True, precision: int = 3, ddof: int = 0
    ) -> str:
        """Returns a string representing the dict of learnable parameters.

        Parameters
        ----------
        summarize : bool, optional
            If `True` (default), array parameters are summarized; otherwise, the entire
            array is printed.
        precision : int, optional
            The printing precision of floating point numbers.
        ddof : int, optional
            Degrees of freedom for computing standard deviations (see `numpy.std`).

        Returns
        -------
        str
            A string representing the dict and its parameters.
        """

        def p2s(p: LearnableParameter) -> str:
            if p.size == 1:
                return f"{p.name}={p.value.item():.{precision}f}"
            if summarize:
                return f"{p.name}: {summarize_array(p.value, precision, ddof)}"
            return np.array2string(p.value, precision=precision)

        return "; ".join(p2s(p) for p in self.values())

    def __str__(self) -> str:
        return self.stringify()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {super().__repr__()}>"



from collections.abc import Iterator
from itertools import chain, cycle, repeat


def bool_cycle(frequency: int, starts_with: bool = False) -> Iterator[bool]:
    """Creates an infinite iterator which cycles via boolean values, where `True`
    appears with the given frequency.

    Parameters
    ----------
    frequency : int
        A positive int specifing the frequency at which `True` appears.
    starts_with : bool, optional
        Whether the first value should be `True` or `False`. By default `False`.

    Returns
    ------
    Iterator of bool
        An iterator with the given frequency of `True`.
    """
    iterator = cycle(chain(repeat(False, frequency - 1), (True,)))
    if starts_with:
        iterator = chain((True,), iterator)
    return iterator



import pickle
import unittest
from collections.abc import Iterator
from copy import deepcopy
from itertools import product
from random import shuffle
from typing import Union
from unittest.mock import Mock

import casadi as cs
import numpy as np
from csnlp.multistart import RandomStartPoints, StructuredStartPoints
from parameterized import parameterized

from mpcrl import (
    ExperienceReplay,
    LearnableParameter,
    LearnableParametersDict,
    MpcSolverError,
    MpcSolverWarning,
    UpdateError,
    UpdateStrategy,
    UpdateWarning,
    WarmStartStrategy,
)
from mpcrl import exploration as E
from mpcrl import schedulers as S
from mpcrl.core.errors import (
    raise_or_warn_on_mpc_failure,
    raise_or_warn_on_update_failure,
)


def do_test_str_and_repr(testcase: unittest.TestCase, obj: S.Scheduler) -> None:
    testcase.assertIn(obj.__class__.__name__, obj.__str__())
    testcase.assertIn(obj.__class__.__name__, obj.__repr__())


def do_test_similar_schedulers(
    testcase: unittest.TestCase, obj1: S.Scheduler, obj2: S.Scheduler
) -> None:
    dict1 = obj1.__dict__
    dict2 = obj2.__dict__
    for name, val1 in dict1.items():
        val2 = dict2[name]
        if hasattr(val1, "__iter__"):
            testcase.assertListEqual(
                [next(val1) for _ in range(10)], [next(val2) for _ in range(10)]
            )
        else:
            testcase.assertEqual(val1, val2)


class TestErrors(unittest.TestCase):
    @parameterized.expand([(False,), (True,)])
    def test_raise_or_warn_on_mpc_failure__raises_or_warns(self, raises: bool):
        msg = "This is a message"
        context = (
            self.assertRaisesRegex(MpcSolverError, msg)
            if raises
            else self.assertWarnsRegex(MpcSolverWarning, msg)
        )
        with context:
            raise_or_warn_on_mpc_failure(msg, raises)

    @parameterized.expand([(False,), (True,)])
    def test_raise_or_warn_on_update_failure__raises_or_warns(self, raises: bool):
        msg = "This is a message"
        context = (
            self.assertRaisesRegex(UpdateError, msg)
            if raises
            else self.assertWarnsRegex(UpdateWarning, msg)
        )
        with context:
            raise_or_warn_on_update_failure(msg, raises)


class TestExperienceReplay(unittest.TestCase):
    def test_init__initializes_memory_correctly(self):
        items = [object(), object(), object()]
        maxlen = 2
        mem = ExperienceReplay[object](items, maxlen=maxlen, seed=50)
        self.assertEqual(mem.maxlen, maxlen)
        self.assertNotIn(items[0], mem)
        self.assertIn(items[1], mem)
        self.assertIn(items[2], mem)
        self.assertIsInstance(mem.np_random, np.random.Generator)

    def test_sample__raises__with_no_maxlen_and_percentage_size(self):
        with self.assertRaises(TypeError):
            ExperienceReplay[tuple[np.ndarray, float]](maxlen=None, sample_size=0.0)

    @parameterized.expand([(0,), (float(0),)])
    def test_sample__with_zero_samples__returns_no_samples(self, n: Union[int, float]):
        mem = ExperienceReplay[tuple[np.ndarray, float]](maxlen=100, sample_size=n)
        self.assertListEqual(list(mem.sample()), [])

    @parameterized.expand([(10,), (0.1,)])
    def test_sample__returns_right_sample_size(self, n: Union[int, float]):
        N = 100
        Nsample = 10
        mem = ExperienceReplay[int](maxlen=N, sample_size=n)
        mem.extend(range(N))
        sample = list(mem.sample())
        self.assertEqual(len(sample), Nsample)
        for item in sample:
            self.assertIn(item, mem)

    @parameterized.expand([(20, 10), (20, 0.5), (0.2, 10), (0.2, 0.5)])
    def test_sample__latest_n__includes_latest_n_items(
        self, n: Union[int, float], last_n: Union[int, float]
    ):
        N = 100
        Nsample = 20
        Nlast = 10
        mem = ExperienceReplay[int](maxlen=N, sample_size=n, include_latest=last_n)
        mem.extend(range(N))
        sample = list(mem.sample())
        self.assertEqual(len(sample), Nsample)
        for item in sample:
            self.assertIn(item, mem)
        for item in range(N - Nlast, N):
            self.assertIn(item, sample)

    def test_deepcopy(self):
        maxlen = np.random.randint(10, 100)
        sample_size, include_latest = np.random.uniform(size=2)
        mem = ExperienceReplay[int](
            maxlen=maxlen, sample_size=sample_size, include_latest=include_latest
        )
        mem_copy = deepcopy(mem)
        self.assertEqual(mem.maxlen, mem_copy.maxlen)
        self.assertEqual(mem.sample_size, mem_copy.sample_size)
        self.assertEqual(mem.include_latest, mem_copy.include_latest)
        self.assertIsInstance(mem_copy.np_random, np.random.Generator)


class TestSchedulers(unittest.TestCase):
    def test_no_scheduling__step__does_not_update_value(self):
        scheduler = S.NoScheduling(init_value=None)
        scheduler.step()
        self.assertIsNone(scheduler.value)
        do_test_str_and_repr(self, scheduler)

    def test_exponential_scheduler__step__updates_value_correctly(self):
        K = 20
        x0 = np.random.randn()
        factor = np.random.rand()
        x_expected = x0 * factor ** np.arange(K)
        scheduler = S.ExponentialScheduler(x0, factor)
        x_actual = []
        for _ in range(K):
            x_actual.append(scheduler.value)
            scheduler.step()
        np.testing.assert_allclose(x_expected, x_actual)
        do_test_str_and_repr(self, scheduler)

    def test_linear_scheduler__step__updates_value_correctly(self):
        # sourcery skip: class-extract-method
        K = 20
        x0 = np.random.rand() * 10
        xf = np.random.rand() * 10
        x_expected = np.linspace(x0, xf, K + 1)
        scheduler = S.LinearScheduler(x0, xf, K)
        x_actual = []
        for _ in range(K):
            x_actual.append(scheduler.value)
            scheduler.step()
        x_actual.append(scheduler.value)
        with self.assertRaises(StopIteration):
            scheduler.step()
        np.testing.assert_allclose(x_expected, x_actual)
        do_test_str_and_repr(self, scheduler)

    def test_loglinear_scheduler__step__updates_value_correctly(self):
        # sourcery skip: class-extract-method
        K = 20
        base = np.random.rand() + 1 * 10
        x0 = np.random.randn() + 3
        xf = np.random.randn()
        x_expected = np.logspace(x0, xf, K + 1, base=base)
        scheduler = S.LogLinearScheduler(base**x0, base**xf, K)
        x_actual = []
        for _ in range(K):
            x_actual.append(scheduler.value)
            scheduler.step()
        x_actual.append(scheduler.value)
        with self.assertRaises(StopIteration):
            scheduler.step()
        np.testing.assert_allclose(x_expected, x_actual)
        do_test_str_and_repr(self, scheduler)

    def test_deepcopy(self):
        K = np.random.randint(10, 20)
        base = np.random.rand() + 1 * 10
        x0 = np.random.randn() + 3
        xf = np.random.randn()
        scheduler = S.LogLinearScheduler(base**x0, base**xf, K)
        scheduler_copy = deepcopy(scheduler)
        do_test_similar_schedulers(self, scheduler, scheduler_copy)

    def test_chain(self):
        # create a list of random schedulers
        rd, ri = np.random.rand, np.random.randint
        mk1 = lambda: (S.NoScheduling(rd()), ri(10, 100))
        mk2 = lambda: (S.ExponentialScheduler(*rd(2)), ri(10, 100))
        mk3 = lambda: S.LinearScheduler(*rd(2), ri(10, 100))
        mk4 = lambda: S.LogLinearScheduler(*rd(2), ri(10, 100))
        schedulers = []
        for mk in (mk1, mk2, mk3, mk4):
            for _ in range(ri(1, 5)):
                schedulers.append(mk())
        shuffle(schedulers)

        # extract the expected values
        values_expected = []
        schedulers_ = deepcopy(schedulers)
        for scheduler in schedulers_:
            if isinstance(scheduler, tuple):
                scheduler, K = scheduler
                while True:
                    values_expected.append(scheduler.value)
                    scheduler.step()
                    K -= 1
                    if K == 0:
                        break
            else:
                while True:
                    values_expected.append(scheduler.value)
                    try:
                        scheduler.step()
                    except StopIteration:
                        break

        # create the chain and extract the actual values
        scheduler = S.Chain(schedulers)
        values_actual = []
        while True:
            values_actual.append(scheduler.value)
            try:
                scheduler.step()
            except StopIteration:
                break

        # test
        steps_expected = sum(
            sc[1] if isinstance(sc, tuple) else sc.total_steps + 1 for sc in schedulers
        )
        self.assertEqual(len(values_actual), steps_expected)
        self.assertEqual(len(values_expected), len(values_actual))
        np.testing.assert_array_equal(values_expected, values_actual)


class TestExploration(unittest.TestCase):
    def test_no_exploration__has_no_mode_nor_hook(self):
        exploration = E.NoExploration()
        self.assertFalse(hasattr(exploration, "_mode"), "should not have `mode`")
        self.assertFalse(hasattr(exploration, "_hook"), "should not have `_hook`")
        self.assertIsNone(exploration.hook)
        self.assertIsNone(exploration.mode)

    def test_no_exploration__never_explores(self):
        exploration = E.NoExploration()
        self.assertFalse(exploration.can_explore())
        with self.assertRaisesRegex(
            NotImplementedError, "Perturbation not implemented in NoExploration"
        ):
            exploration.perturbation()
        exploration.step()  # does nothing
        do_test_str_and_repr(self, exploration)

    def test_greedy_exploration__instantiates_np_random(self):
        exploration = E.GreedyExploration(strength=0.5)
        self.assertIsInstance(exploration.np_random, np.random.Generator)

    def test_greedy_exploration__always_explores(self):
        exploration = E.GreedyExploration(strength=0.5)
        self.assertTrue(exploration.can_explore())
        do_test_str_and_repr(self, exploration)

    @parameterized.expand(((False,), (True,)))
    def test_greedy_exploration__hook(self, strength):
        strength_scheduler = S.LinearScheduler(0, 1, 10) if strength else 0
        exploration = E.GreedyExploration(strength_scheduler)
        hook = exploration.hook
        self.assertEqual(hook is not None, strength)

    @parameterized.expand([("uniform",), ("normal",), ("standard_normal",)])
    def test_greedy_exploration__perturbs(self, method: str):
        exploration = E.GreedyExploration(strength=0.5)
        exploration.perturbation(method)

    def test_epsilon_greedy_exploration__never_explores__with_zero_epsilon(self):
        epsilon, epsilon_decay_rate = 0.0, 0.75
        strength, strength_decay_rate = 0.5, 0.75
        epsilon_scheduler = S.ExponentialScheduler(epsilon, epsilon_decay_rate)
        strength_scheduler = S.ExponentialScheduler(strength, strength_decay_rate)
        exploration = E.EpsilonGreedyExploration(
            epsilon=epsilon_scheduler, strength=strength_scheduler, seed=42
        )
        self.assertFalse(any(exploration.can_explore() for _ in range(100)))

    def test_epsilon_greedy_exploration__sometimes_explores(self):
        epsilon, epsilon_decay_rate = 0.7, 0.75
        strength, strength_decay_rate = 0.5, 0.75
        epsilon_scheduler = S.ExponentialScheduler(epsilon, epsilon_decay_rate)
        strength_scheduler = S.ExponentialScheduler(strength, strength_decay_rate)
        exploration = E.EpsilonGreedyExploration(
            epsilon=epsilon_scheduler, strength=strength_scheduler, seed=42
        )
        self.assertFalse(exploration.can_explore())
        found_true, found_false = False, False
        for _ in range(100):
            explore = exploration.can_explore()
            if explore:
                found_true = True
            else:
                found_false = True
            if found_true and found_false:
                break
        self.assertTrue(found_true and found_false)

    def test_epsilon_greedy_exploration__decays_strength(self):
        class MockScheduler(S.NoScheduling):
            ...

        epsilon_scheduler = MockScheduler(None)
        strength_scheduler = MockScheduler(None)
        epsilon_scheduler.step = Mock()
        strength_scheduler.step = Mock()
        exploration = E.EpsilonGreedyExploration(
            epsilon=epsilon_scheduler, strength=strength_scheduler, seed=42
        )

        exploration.step()

        epsilon_scheduler.step.assert_called_once()
        strength_scheduler.step.assert_called_once()

    @parameterized.expand(product([False, True], [False, True]))
    def test_epsilon_greedy_exploration__hook(self, epsilon, strength):
        epsilon_scheduler = S.LinearScheduler(0, 1, 10) if epsilon else 0
        strength_scheduler = S.LinearScheduler(0, 1, 10) if strength else 0
        exploration = E.EpsilonGreedyExploration(epsilon_scheduler, strength_scheduler)
        hook = exploration.hook
        self.assertEqual(hook is not None, epsilon or strength)

    def test_ornsteinuhlenbeck_exploration__always_explores(self):
        exploration = E.OrnsteinUhlenbeckExploration(0, 0.5)
        self.assertTrue(exploration.can_explore())

    def test_ornsteinuhlenbeck_exploration__decays_mean_and_sigma(self):
        class MockScheduler(S.NoScheduling):
            ...

        mean_scheduler = MockScheduler(None)
        sigma_scheduler = MockScheduler(None)
        mean_scheduler.step = Mock()
        sigma_scheduler.step = Mock()
        exploration = E.OrnsteinUhlenbeckExploration(mean_scheduler, sigma_scheduler)

        exploration.step()

        mean_scheduler.step.assert_called_once()
        sigma_scheduler.step.assert_called_once()

    @parameterized.expand(product([False, True], [False, True]))
    def test_ornsteinuhlenbeck_exploration__hook(self, mean, sigma):
        mean_scheduler = S.LinearScheduler(0, 1, 10) if mean else 0
        sigma_scheduler = S.LinearScheduler(0, 1, 10) if sigma else 0
        exploration = E.OrnsteinUhlenbeckExploration(mean_scheduler, sigma_scheduler)
        hook = exploration.hook
        self.assertEqual(hook is not None, mean or sigma)

    def test_stepwise_exploration__has_same_hook_and_mode_as_base_exploration(self):
        hook, mode = Mock(), Mock()
        base_exploration = Mock()
        base_exploration.hook = hook
        base_exploration.mode = mode
        exploration = E.StepWiseExploration(base_exploration, 5, 10)
        self.assertIs(exploration.hook, hook)
        self.assertIs(exploration.mode, mode)

    @parameterized.expand([(True,), (False,)])
    def test_stepwise_exploration__turns_base_exploration_into_steps(
        self, stepwise_step: bool
    ):
        base_exploration = E.EpsilonGreedyExploration(0.5, 0.5, seed=0)
        base_exploration.step = Mock()
        step_size = 5
        cycles = 10
        exploration = E.StepWiseExploration(base_exploration, step_size, stepwise_step)
        can_explores, perturbations = [], []
        for _ in range(cycles):
            can_explores.append([])
            perturbations.append([])
            for _ in range(step_size):
                can_explores[-1].append(exploration.can_explore())
                perturbations[-1].append(exploration.perturbation("uniform"))
                exploration.step()

        for can_explores_cycle, perturbations_cycle in zip(can_explores, perturbations):
            self.assertTrue(np.unique(can_explores_cycle).size == 1)
            self.assertTrue(np.unique(perturbations_cycle).size == 1)
        if stepwise_step:
            self.assertTrue(len(base_exploration.step.mock_calls) == cycles)
        else:
            self.assertTrue(len(base_exploration.step.mock_calls) == cycles * step_size)

    def test_deepcopy(self):
        epsilon, epsilon_decay_rate = 0.0, 0.75
        strength, strength_decay_rate = 0.5, 0.75
        epsilon_scheduler = S.ExponentialScheduler(epsilon, epsilon_decay_rate)
        strength_scheduler = S.ExponentialScheduler(strength, strength_decay_rate)
        exploration = E.EpsilonGreedyExploration(
            epsilon=epsilon_scheduler, strength=strength_scheduler, seed=42
        )
        exploration_copy = deepcopy(exploration)
        self.assertEqual(exploration._hook, exploration_copy._hook)
        do_test_similar_schedulers(
            self, exploration.epsilon_scheduler, exploration_copy.epsilon_scheduler
        )
        do_test_similar_schedulers(
            self, exploration.strength_scheduler, exploration_copy.strength_scheduler
        )


class TestParameters(unittest.TestCase):
    @parameterized.expand(map(lambda i: (i,), range(3)))
    def test_learnable_parameter_init__raises__with_unbroadcastable_args(self, i: int):
        args = [5, 0, 10]
        args[i] = np.random.rand(2, 2, 2)
        with self.assertRaises(ValueError):
            LearnableParameter("theta", 10, *args)

    def test_learnable_parameter_init_and_update__raise__with_value_outside_range(self):
        with self.assertRaises(ValueError):
            LearnableParameter("theta", 10, -5, 0, 10)
        par = LearnableParameter("theta", 10, 5, 0, 10)
        with self.assertRaises(ValueError):
            par._update_value(-5)

    @parameterized.expand([(False,), (True,)])
    def test_parameters_dict__init__initializes_properly(self, empty_init: bool):
        p1 = LearnableParameter("theta1", 10, 0.0)
        p2 = LearnableParameter("theta2", 11, 0.0)
        PARS = [p1, p2]
        if empty_init:
            pars = LearnableParametersDict()
            pars.update(iter(PARS))
        else:
            pars = LearnableParametersDict(iter(PARS))
        self.assertEqual(len(pars), 2)
        for p in PARS:
            self.assertIn(p.name, pars)
            self.assertIs(pars[p.name], p)

    @parameterized.expand([(False,), (True,)])
    def test_parameters_dict__setitem__raises_with_wrong_name(self, wrong_name: bool):
        p = LearnableParameter("t", 10, 0.0)
        pars = LearnableParametersDict()
        if wrong_name:
            with self.assertRaisesRegex(
                AssertionError, r"Key 't_hello_there' must match parameter name 't'."
            ):
                pars[f"{p.name}_hello_there"] = p
        else:
            pars[p.name] = p

    @parameterized.expand(
        map(
            lambda m: (m,),
            ["setitem", "update", "setdefault", "delitem", "pop", "popitem", "clear"],
        )
    )
    def test_parameters_dict__caches_get_cleared_properly(self, method: str):
        array_equal = np.testing.assert_array_equal
        cat = np.concatenate

        def check(pars, PARS):
            self.assertEqual(pars.size, sum(p.size for p in PARS))
            if len(pars) > 0 and len(PARS) > 0:
                array_equal(pars.lb, cat([p.lb.flatten("F") for p in PARS]))
                array_equal(pars.ub, cat([p.ub.flatten("F") for p in PARS]))
                array_equal(pars.value, cat([p.value.flatten("F") for p in PARS]))
                sym1 = cs.veccat(*(p.sym for p in pars.values()))
                sym2 = cs.veccat(*(p.sym for p in PARS))
                array_equal(cs.evalf(cs.simplify(sym1 - sym2)), 0)
            else:
                self.assertTupleEqual(pars.lb.shape, (0,))
                self.assertTupleEqual(pars.ub.shape, (0,))
                self.assertTupleEqual(pars.value.shape, (0,))

        p1 = LearnableParameter[float](
            "t1", (4, 6), 1.0, -1.0, 2.0, cs.SX.sym("t1", 4, 6)
        )
        p2 = LearnableParameter[float]("t2", 2, 2.0, -2.0, 3.0, cs.SX.sym("t2", 2))
        PARS = [p1, p2]
        pars = LearnableParametersDict(PARS)
        check(pars, PARS)

        p3 = LearnableParameter[float]("t3", 3, 3.0, -3.0, 4.0, cs.SX.sym("t3", 3))
        if method == "setitem":
            pars[p3.name] = p3
            PARS = [p1, p2, p3]
        elif method == "update":
            pars.update([p3])
            PARS = [p1, p2, p3]
        elif method == "setdefault":
            pars.setdefault(p3)
            PARS = [p1, p2, p3]
        elif method == "delitem":
            del pars[p2.name]
            PARS = [p1]
        elif method == "pop":
            pars.pop(p1.name)
            PARS = [p2]
        elif method == "popitem":
            pars.popitem()
            PARS = [p1]
        elif method == "clear":
            pars.clear()
            PARS = []

        check(pars, PARS)

    def test_parameters_dict__stringify(self):
        p1 = LearnableParameter[None]("p1", 1, 1)
        p2 = LearnableParameter[None]("p2", 1, 2.0)
        p3 = LearnableParameter[None]("p3", 100, np.random.randint(0, 100, size=100))
        p4 = LearnableParameter[None]("p4", 100, np.random.rand(100))
        pars = LearnableParametersDict((p1, p2, p3, p4))
        S = pars.stringify()
        for p in [p1, p2, p3, p4]:
            self.assertIn(p.name, S)

    @parameterized.expand(product([cs.SX, cs.MX], [False, True]))
    def test_deepcopy(self, cstype: Union[cs.SX, cs.MX], copy: bool):
        shape = (5, 2)
        theta_sym = cstype.sym("theta", shape)
        f = theta_sym[0]
        df = cs.evalf(cs.jacobian(f, theta_sym)[0])
        p1 = LearnableParameter[float]("theta", shape, 1, -1, 2, sym={"v": theta_sym})
        pars = LearnableParametersDict((p1,))
        if copy:
            new_pars = pars.copy(deep=True)
        else:
            new_pars: LearnableParametersDict = pickle.loads(pickle.dumps(pars))
        p2: LearnableParameter = new_pars["theta"]
        self.assertIsNot(p1, p2)
        self.assertEqual(p1.name, p2.name)
        self.assertEqual(p1.shape, p2.shape)
        self.assertEqual(p1.size, p2.size)
        np.testing.assert_array_equal(p1.value, p2.value)
        np.testing.assert_array_equal(p1.lb, p2.lb)
        np.testing.assert_array_equal(p1.ub, p2.ub)
        np.testing.assert_equal(df, cs.evalf(cs.jacobian(f, p1.sym["v"])[0]))
        if copy:
            self.assertIsNot(p1.sym, p2.sym)
            self.assertIsNot(p1.sym["v"], p2.sym["v"])
            np.testing.assert_equal(df, cs.evalf(cs.jacobian(f, p2.sym["v"])[0]))
        else:
            self.assertFalse(hasattr(p2, "sym"))


class TestUpdateStrategy(unittest.TestCase):
    def test__is_iterable(self):
        strategy = UpdateStrategy(5)
        self.assertIsInstance(next(strategy), bool)
        self.assertIsInstance(iter(strategy), Iterator)

    def test_can_update__has_right_frequency(self):
        N = 50
        freq = 3
        strategy = UpdateStrategy(freq)
        flags = np.asarray([strategy.can_update() for _ in range(N)])
        mask = np.asarray(([False] * (freq - 1) + [True]) * (N // freq + 1))[:N]
        self.assertEqual(flags.sum(), N // freq)
        self.assertTrue(flags[mask].all())
        self.assertTrue((~flags[~mask]).all())

    def test_can_update__with_skip_first__skips_first_updates_correctly(self):
        N = 50
        freq = 3
        skip = 2
        strategy = UpdateStrategy(freq, skip_first=skip)
        flags = np.asarray([strategy.can_update() for _ in range(N)])
        mask = np.asarray(([False] * (freq - 1) + [True]) * (N // freq + 1))[:N]
        self.assertEqual(flags.sum(), N // freq - skip)
        self.assertTrue((~flags[mask][:skip]).all())
        self.assertTrue(flags[mask][skip:].all())
        self.assertTrue((~flags[~mask]).all())

    def test_deepcopy(self):
        strategy = UpdateStrategy(5, "on_episode_end", 3)
        strategy_copy = deepcopy(strategy)
        self.assertEqual(strategy.frequency, strategy_copy.frequency)
        self.assertEqual(strategy.hook, strategy_copy.hook)
        self.assertListEqual(
            [next(strategy._update_cycle) for _ in range(50)],
            [next(strategy_copy._update_cycle) for _ in range(50)],
        )


class TestWarmStartStrategy(unittest.TestCase):
    @parameterized.expand(product((False, True), (False, True)))
    def test_n_points(self, struct_points: bool, rand_points: bool):
        ns = np.random.randint(3, 100) if struct_points else 0
        nr = np.random.randint(3, 100) if rand_points else 0
        wss = WarmStartStrategy(
            structured_points=StructuredStartPoints({}, ns) if struct_points else None,
            random_points=RandomStartPoints({}, nr) if rand_points else None,
        )
        self.assertEqual(wss.n_points, ns + nr)

    def test_reset(self):
        wss = WarmStartStrategy()
        random_points = Mock()
        wss.random_points = random_points
        wss.reset(42)
        self.assertIsInstance(random_points.np_random, np.random.Generator)

    def test_generate(self):
        struct_points = Mock()
        rand_points = Mock()
        struct_points.__iter__ = lambda self: iter((self,))
        rand_points.__iter__ = lambda self: iter((self,))
        wss = WarmStartStrategy(
            structured_points=struct_points,
            random_points=rand_points,
            update_biases_for_random_points=True,
        )
        biases = object()
        out = list(wss.generate(biases=biases))
        self.assertIs(out[0], struct_points)
        self.assertIs(out[1], rand_points)
        rand_points.biases.update.assert_called_once_with(biases)


if __name__ == "__main__":
    unittest.main()



from itertools import combinations
from typing import Union

import casadi as cs
import numpy as np
import numpy.typing as npt
from csnlp.util.math import prod
from scipy.special import comb


def summarize_array(
    a: npt.NDArray[np.floating], precision: int = 3, ddof: int = 0
) -> str:
    """Summarizes the stats of a given array.

    Parameters
    ----------
    a : array
        The numerical array to summarize.
    precision : int, optional
        The decimal precision, by default 3.
    ddof : int, optional
        Degrees of freedom used to compute the standard deviation, by default 0.

    Returns
    -------
    str
        The summarizing string.
    """
    s = a.shape
    mean = a.mean()
    std = a.std(ddof=ddof)
    min = a.min()
    max = a.max()
    return (
        f"s={s} x∈[{min:.{precision}f}, {max:.{precision}f}] "
        f"μ={mean:.{precision}f} σ={std:.{precision}f}"
    )


def cholesky_added_multiple_identities(
    A: npt.NDArray[np.floating], beta: float = 1e-3, maxiter: int = 1000
) -> npt.NDArray[np.floating]:
    """Lower Cholesky factorization with added multiple of the identity matrix to ensure
    positive-definitiveness [1, Algorithm 3.3].

    Parameters
    ----------
    A : array of double
        The matrix to compute the cholesky factorization of.
    beta : float, optional
        Initial tolerance of the algorithm, by default 1e-3.
    maxiter : int, optional
        Maximum iterations of the algorithm, by default 1000.

    Returns
    -------
    array of double
        The lower cholesky factorization of the modified A (with the addition of
        identity matrices to ensure that it is positive-definite).

    Raises
    ------
    ValueError
        If the factorization is unsuccessful for the maximum number of iterations.

    References
    ----------
    [1] J. Nocedal, S. Wright. 2006. Numerical Optimization. Springer.
    """
    a_min = np.diag(A).min()
    tau = 0 if a_min > 0 else -a_min + beta
    identity = np.eye(A.shape[0])
    for _ in range(maxiter):
        try:
            return np.linalg.cholesky(A + tau * identity)
        except np.linalg.LinAlgError:
            tau = max(1.05 * tau, beta)
    raise ValueError("Maximum iterations reached.")


def nchoosek(n: Union[int, npt.ArrayLike], k: int) -> Union[int, np.ndarray]:
    """Emulates the `nchoosek` function from Matlab. Returns the binomial coefficient,
    i.e.,  the number of combinations of `n` items taken `k` at a time. If `n` is an
    array, then it is flatten and all possible combinations of its elements are
    returned.

    Parameters
    ----------
    n : int or array_like
        Number of elements or array of elements to choose from.
    k : int
        Number of elements to choose.

    Returns
    -------
    int or array
        Depending on the type of input `n`, the output is either the total number of
        combinations or the combinations in a matrix.
    """
    return (
        comb(n, k, True)
        if isinstance(n, int)
        else np.row_stack(list(combinations(np.asarray(n).reshape(-1), k)))
    )


def monomial_powers(d: int, k: int) -> npt.NDArray[np.int64]:
    """Computes the powers of all `d`-dimensional monomials of degree `k`.

    Parameters
    ----------
    d : int
        The number of monomial elements.
    k : int
        The degree of each monomial.

    Returns
    -------
    array of ints
        An array containing in each row the power of each index in order to obtain the
        desired monomial of power `k`.
    """
    m = nchoosek(k + d - 1, d - 1)
    dividers = np.column_stack(
        (
            np.zeros((m, 1), int),
            np.row_stack(nchoosek(np.arange(1, k + d), d - 1)),
            np.full((m, 1), k + d, int),
        )
    )
    return np.flipud(np.diff(dividers, axis=1) - 1).astype(int)


def monomials_basis_function(n_in: int, mindegree: int, maxdegree: int) -> cs.Function:
    """Creates a function made of monomials as bases.

    Parameters
    ----------
    ns : int
        Number of states.
    mindegree : int
        Minimum degree of monomials (included).
    maxdegree : int
        Maximum degree of monomials (included).

    Returns
    -------
    casadi.Function
        A casadi function of the form `Phi(s)`, where `s` is the input and `Phi` the
        basis vector.
    """
    s = cs.SX.sym("x", n_in, 1)  # prod is faster with SX, and during runtime
    y = cs.vertcat(
        *(
            prod(s**p)
            for k in range(mindegree, maxdegree + 1)
            for p in monomial_powers(n_in, k)
        )
    )
    return cs.Function("Phi", (s,), (y,), ("s",), ("Phi(s)",), {"cse": True})



import unittest

import casadi as cs
import numpy as np
from parameterized import parameterized, parameterized_class

from mpcrl.util import control, iters, math, named, seeding


class DummyAgent(named.Named):
    ...


class TestNamedAgent(unittest.TestCase):
    def test_init__with_given_name__saves_name_correctly(self):
        dummy = DummyAgent(name="ciao")
        self.assertEqual(dummy.name, "ciao")

    def test_init__without_name__creates_name_with_class(self):
        dummy0 = DummyAgent()
        dummy1 = DummyAgent()
        self.assertEqual(dummy0.name, "DummyAgent0")
        self.assertEqual(dummy1.name, "DummyAgent1")

    def test_str_and_repr(self):
        name = "ciao"
        dummy = DummyAgent(name=name)
        S = dummy.__str__()
        self.assertIn(name, S)
        S = dummy.__repr__()
        self.assertIn(name, S)
        self.assertIn(DummyAgent.__name__, S)


class TestMath(unittest.TestCase):
    def test_cholesky_added_multiple_identities__performs_correct_factorization(self):
        n = 5
        A = np.random.rand(n, n) + np.eye(n) * 5
        A = 0.5 * (A + A.T)
        L = math.cholesky_added_multiple_identities(A)
        np.testing.assert_allclose(A, L @ L.T)

    def test_cholesky_added_multiple_identities__raises__max_iterations_reached(self):
        n = 5
        A = np.random.rand(n, n) - np.eye(n) * 5
        A = 0.5 * (A + A.T)
        with self.assertRaisesRegex(ValueError, "Maximum iterations reached."):
            math.cholesky_added_multiple_identities(A, maxiter=1)

    @parameterized.expand(
        [
            ((1, 1), 1),
            ((5, 4), 5),
            (
                (np.arange(2, 11, 2), 4),
                np.asarray(
                    [
                        [2, 4, 6, 8],
                        [2, 4, 6, 10],
                        [2, 4, 8, 10],
                        [2, 6, 8, 10],
                        [4, 6, 8, 10],
                    ]
                ),
            ),
            ((np.asarray([10, 20, 30]), 2), np.asarray([[10, 20], [10, 30], [20, 30]])),
        ]
    )
    def test_nchoosek__computes_correct_combinations(
        self, inp: tuple[int, int], out: int
    ):
        out_ = math.nchoosek(*inp)
        np.testing.assert_allclose(out_, out)

    @parameterized.expand([(1, 4, 4), (10, 1, np.eye(10)), (4, 3, None)])
    def test_monomial_powers__computes_correct_powers(self, n, k, out):
        p = math.monomial_powers(n, k)
        self.assertEqual(p.shape[1], n)
        np.testing.assert_allclose(p.sum(axis=1), k)
        if out is not None:
            np.testing.assert_allclose(p, out)


class TestControl(unittest.TestCase):
    def test_dlqr__returns_correctly(self):
        K_exp = np.array([[1.075936290787970, 1.824593914133278]])
        P_exp = np.array(
            [
                [6.783278637425703, 2.903698804441288],
                [2.903698804441288, 5.026668672843932],
            ]
        )
        A = np.array([[1, 0.25], [0, 1]])
        B = np.array([[0.03], [0.25]])
        Q = np.eye(2)
        R = 0.5
        K, P = control.dlqr(A, B, Q, R)
        np.testing.assert_allclose(K, K_exp)
        np.testing.assert_allclose(P, P_exp)

    def test_dlqr__with_M__returns_correctly(self):
        K_exp = np.array([[1.132928272226698, 1.820247185260309]])
        P_exp = np.array(
            [
                [6.426698776552665, 2.359469798498852],
                [2.359469798498852, 3.806839220681791],
            ]
        )
        A = np.array([[1, 0.25], [0, 1]])
        B = np.array([[0.03], [0.25]])
        Q = np.eye(2)
        R = np.atleast_2d(0.5)
        M = np.array([[0.1], [0.2]])
        K, P = control.dlqr(A, B, Q, R, M)
        np.testing.assert_allclose(K, K_exp)
        np.testing.assert_allclose(P, P_exp)

    def test_rk4__returns_correcly(self):
        def f(x):
            return -3 * x

        dt = 0.01
        x = cs.SX.sym("x")
        fd = cs.Function("fd", [x], [control.rk4(f, x, dt)])

        Y = [1]
        for _ in range(100):
            Y.append(fd(Y[-1]))
        Y = cs.hcat(Y).full().squeeze()
        Y_exact = np.exp(-3 * (np.arange(Y.size) * dt))
        np.testing.assert_allclose(Y, Y_exact)


@parameterized_class("starts_with", [(False,), (True,)])
class TestIters(unittest.TestCase):
    @parameterized.expand([(5,), (1,), (22,)])
    def test_bool_cycle(self, frequency: int):
        T = frequency * 10
        cycle = iters.bool_cycle(frequency, self.starts_with)
        cycle = [next(cycle) for _ in range(T)]
        self.assertEqual(cycle[0], self.starts_with or frequency == 1)
        self.assertEqual(T // frequency, sum(cycle))


class TestSeeding(unittest.TestCase):
    def test_mk_seed(self):
        rng = np.random.default_rng()
        self.assertTrue(0 <= seeding.mk_seed(rng) < 2**32)


if __name__ == "__main__":
    unittest.main()



from collections.abc import Sequence
from typing import Union

import numpy as np
from typing_extensions import TypeAlias

RngType: TypeAlias = Union[
    None,
    int,
    Sequence[int],
    np.random.SeedSequence,
    np.random.BitGenerator,
    np.random.Generator,
]


MAX_SEED = np.iinfo(np.uint32).max + 1


def mk_seed(rng: np.random.Generator) -> int:
    """Generates a random seed.

    Parameters
    ----------
    rng : np.random.Generator
        RNG generator

    Returns
    -------
    int
        A random integer in the range [0, 2**32)
    """
    return int(rng.integers(MAX_SEED))
